# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard library imports
import time
import ast
from functools import reduce
from threading import Lock
import logging
from multiprocessing.managers import ListProxy

# Third party imports

# Local application imports
from .image_writer import ImageWriter
from navigate.tools.common_functions import VariableWithLock


from navigate.model.waveforms import remote_focus_ramp

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class Snap:
    """Snap class for capturing data frames using a microscope.

    This class provides functionality to capture data frames using a microscope
    and log information about the camera and frame IDs.

    Notes:
    ------
    - This class is used to capture data frames using a microscope and log
    relevant information, such as the active camera and frame IDs.

    - The `Snap` class is typically used for capturing individual frames during
    microscopy experiments.

    - The data capture process involves capturing frames and logging
    camera-related information.

    - The `config_table` attribute is used to define the configuration for the
    data capture process, specifically the main data capture function.
    """

    def __init__(self, model, saving_flag=False):
        """Initialize the Snap class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for data capture.
        """
        #: MicroscopeModel: The microscope model associated with the data capture.
        self.model = model

        #: bool: Saving each frames
        self.saving_flag = saving_flag

        #: dict: A dictionary defining the configuration for the data capture process.
        self.config_table = {
            "signal": {"main": self.signal_func},
            "data": {"main": self.data_func},
        }

    def signal_func(self) -> bool:
        """Mark saving flags in the signal function"""
        if self.saving_flag:
            self.model.mark_saving_flags([self.model.frame_id])
        return True

    def data_func(self, frame_ids: list) -> bool:
        """Capture data frames and log camera information.

        This method captures data frames using the microscope and logs information
        about the active camera and the provided frame IDs.

        Parameters:
        ----------
        frame_ids : list
            A list of frame IDs for which data frames should be captured.

        Returns:
        -------
        bool
            A boolean value indicating the success of the data capture process.
        """
        logger.info(f"the camera is:{self.model.active_microscope_name}, {frame_ids}")
        return True


class WaitForExternalTrigger:
    """WaitForExternalTrigger class to time features using external input.

    This class waits for either an external trigger (or the timeout) before continuing
    on to the next feature block in the list. Useful when combined with LoopByCount
    when each iteration may depend on some external event happening.

    Notes:
    ------
    - This class pauses the data thread while waiting for the trigger to avoid
      camera timeout issues.

    - Only digital triggers are handeled at this time: use the PFI inputs on the DAQ.
    """

    def __init__(self, model, trigger_channel="/PCIe-6738/PFI4", timeout=-1):
        """Initialize the WaitForExternalTrigger class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for synchronization.
        trigger_channel : str
            The name of the DAQ PFI digital input.
        timeout : float
            Continue on anyway if timeout is reached. timeout < 0 will
            run forever.
        """
        self.model = model

        self.wait_interval = 0.001  # sec

        self.task = None
        self.trigger_channel = trigger_channel
        self.timeout = timeout

        self.config_table = {
            "signal": {
                "main": self.signal_func,
            }
        }

    def signal_func(self):

        # Pause the data thread to prevent camera timeout
        self.model.pause_data_thread()

        result = self.model.active_microscope.daq.wait_for_external_trigger(
            self.trigger_channel, self.wait_interval, self.timeout
        )

        # Resume the data thread
        self.model.resume_data_thread()

        return result

class ProjectionMode:

    def __init__(self, model, axis='z', galvo_num=0, enable=True, z_range=None, shear_amp=None):

        self.model = model

        self.enable = enable

        self.z_range = z_range
        self.shear_amp = shear_amp

        self.microscope_state = None
        self.waveform_constants = None

        self.galvo_stage = model.active_microscope.stages[axis]
        self.shear_galvo = model.active_microscope.galvo[f"galvo_{galvo_num}"]

        self.galvo_num = galvo_num

        self.exposure_times = None
        self.sweep_times = None
        self.channels = None

        self.current_channel_in_list = 0

        self.config_table = {
            "signal": {"main": self.toggle_projection_mode}
        }

    def toggle_projection_mode(self):

        self.microscope_state = self.model.configuration["experiment"]["MicroscopeState"]
        self.waveform_constants = self.model.configuration["waveform_constants"]

        (
            self.exposure_times,
            self.sweep_times
        ) = self.model.active_microscope.get_exposure_sweep_times()

        if self.enable:
            self.setup_projection()
        else:
            self.disable_projection()

        return True

    def setup_projection(self):

        self.microscope_state["waveform_template"] = "Confocal-Projection"

        self.channels = self.microscope_state["selected_channels"]

        remote_focus_delay = float(self.waveform_constants["other_constants"][
            "remote_focus_delay"
            ]) / 1000
        remote_focus_ramp_falling = float(self.waveform_constants["other_constants"][
            "remote_focus_ramp_falling"
            ]) / 1000
        
        z_range = self.microscope_state["scanrange"] if self.z_range is None else self.z_range        
        shear_amp = self.microscope_state["shear_amp"] if self.shear_amp is None else self.shear_amp

        waveform_dict = {}
        for channel_key in self.microscope_state["channels"].keys():
            
            channel = self.microscope_state["channels"][channel_key]

            if channel["is_selected"]:

                waveform_dict[channel_key] = remote_focus_ramp(
                    sample_rate = self.galvo_stage.sample_rate,
                    exposure_time = self.exposure_times[channel_key],
                    sweep_time = self.sweep_times[channel_key],
                    remote_focus_delay = remote_focus_delay,
                    fall = remote_focus_ramp_falling,
                    camera_delay = self.galvo_stage.camera_delay,
                    amplitude = eval(self.galvo_stage.volts_per_micron, {"x": 0.5 * (z_range)})
                )
        
        self.galvo_stage.update_waveform(waveform_dict)

        self.set_shear_amplitude(shear_amp)

    def set_shear_amplitude(self, amp):
            
        self.waveform_constants["galvo_constants"][f"Galvo {self.galvo_num}"][
        self.microscope_state["microscope_name"]
        ][
            self.microscope_state["zoom"]
        ]["amplitude"] = amp

        self.shear_galvo.adjust(
            self.exposure_times,
            self.sweep_times
        )

    def disable_projection(self):

        self.microscope_state["waveform_template"] = "Default"        

        self.galvo_stage.waveform_dict = {}

        self.set_shear_amplitude(0)

        # self.galvo_stage.ao_task = None

        self.galvo_stage.switch_mode(
            "normal",
            # exposure_times = self.exposure_times,
            # sweep_times = self.sweep_times
            )

class WaitToContinue:
    """WaitToContinue class for synchronizing signal and data acquisition.

    This feature is used to synchronize signal and data acquisition processes, allowing
    the faster one to wait until the other one ends.

    Notes:
    ------
    - This class is used to synchronize signal and data acquisition processes in a
    controlled manner. It ensures that the faster process waits for the slower one to
    complete, improving synchronization during microscopy experiments.

    - The synchronization process involves using locks to control when the signal and
    data acquisition processes are allowed to proceed.

    - The `config_table` attribute defines the configuration for each stage of the
    synchronization process, including initialization, main execution, and cleanup
    steps.

    - The `first_enter_node` attribute tracks which process (signal or data) enters the
    node first to determine synchronization order.

    - The `WaitToContinue` class helps maintain order and synchronization between signal
    and data acquisition nodes in microscopy experiments.
    """

    def __init__(self, model):
        """Initialize the WaitToContinue class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for synchronization.
        """
        #: MicroscopeModel: The microscope model associated with the synchronization.
        self.model = model

        #: Lock: A lock for the signal acquisition process.
        self.pause_signal_lock = Lock()

        #: Lock: A lock for the data acquisition process.
        self.pause_data_lock = Lock()

        #: VariableWithLock: A variable to track which process enters the node first.
        self.first_enter_node = VariableWithLock(str)

        #: dict: A dictionary defining the configuration for the synchronization
        self.config_table = {
            "signal": {
                "init": self.pre_signal_func,
                "main": self.signal_func,
                "cleanup": self.cleanup,
            },
            "data": {
                "init": self.pre_data_func,
                "main": self.data_func,
                "cleanup": self.cleanup,
            },
        }

    def pre_signal_func(self):
        """Prepare for the signal acquisition stage and synchronize with data
        acquisition.

        This method prepares for the signal acquisition stage and synchronizes with
        the data acquisition process, ensuring that the slower process proceeds first.
        """
        with self.first_enter_node as first_enter_node:
            if first_enter_node.value == "":
                logger.debug("*** wait to continue enters signal " "first!")
                first_enter_node.value = "signal"
                if not self.pause_signal_lock.locked():
                    self.pause_signal_lock.acquire()
                if self.pause_data_lock.locked():
                    self.pause_data_lock.release()

    def signal_func(self):
        """Synchronize signal acquisition and release locks.

        This method synchronizes the signal acquisition process with data acquisition
        and releases any locks held.

        Returns:
        -------
        bool
           A boolean value indicating the success of the synchronization process.
        """
        logger.debug(f"--wait to continue: {self.model.frame_id}")
        if self.pause_signal_lock.locked():
            self.pause_signal_lock.acquire()
        elif self.pause_data_lock.locked():
            self.pause_data_lock.release()
        self.first_enter_node.value = ""
        logger.debug(f"--wait to continue is done!: {self.model.frame_id}")
        return True

    def pre_data_func(self):
        """Prepare for the data acquisition stage and synchronize with signal
        acquisition.

        This method prepares for the data acquisition stage and synchronizes with the
        signal acquisition process, ensuring that the slower process proceeds first.
        """
        with self.first_enter_node as first_enter_node:
            if first_enter_node.value == "":
                logger.debug("*** wait to continue enters data " "node first!")
                first_enter_node.value = "data"
                if not self.pause_data_lock.locked():
                    self.pause_data_lock.acquire()
                if self.pause_signal_lock.locked():
                    self.pause_signal_lock.release()

    def data_func(self, frame_ids):
        """Synchronize data acquisition and release locks.

        This method synchronizes the data acquisition process with signal acquisition
        and releases any locks held.

        Parameters:
        ----------
        frame_ids : list
            A list of frame IDs for which data acquisition should be performed.

        Returns:
        -------
        bool
            A boolean value indicating the success of the synchronization process.
        """
        logger.debug(f"**wait to continue? {frame_ids}")
        if self.pause_data_lock.locked():
            self.pause_data_lock.acquire()
        elif self.pause_signal_lock.locked():
            self.pause_signal_lock.release()
        self.first_enter_node.value = ""
        logger.debug(f"**wait to continue is done! {frame_ids}")
        return True

    def cleanup(self):
        """Release any remaining locks during cleanup.

        This method releases any locks that may still be held during cleanup.
        """
        if self.pause_signal_lock.locked():
            self.pause_signal_lock.release()
        if self.pause_data_lock.locked():
            self.pause_data_lock.release()


class LoopByCount:
    """LoopByCount class for controlling signal and data acquisition loops.

    This class provides functionality to control signal and data acquisition loops by
    specifying the number of steps or frames to execute.

    Notes:
    ------
    - This class is used to control signal and data acquisition loops by specifying the
    number of steps or frames to execute. It allows for flexible control of the
    acquisition process.

    - The `steps` parameter can be an integer specifying the number of steps/frames
    directly or a string representing a configuration reference to determine the loop
    count dynamically.

    - The loop control process involves tracking the remaining steps/frames and deciding
     whether to continue the loop or exit based on the remaining count.

    - The `LoopByCount` class is useful for controlling the number of acquisitions
    during microscopy experiments, either by specifying a fixed count or by
    dynamically determining it from configuration references.
    """

    def __init__(self, model, steps=1, step_by_frame=False, is_nested=False):
        """Initialize the LoopByCount class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for loop control.
        steps : int or str, optional
            The number of steps or a configuration reference to determine the loop
            count. Default is 1.
        step_by_frame : bool
            Count by number of frames received/ the number of entering this node.
        is_nested : bool
            The flag indicates whether the loop is nested within another loop.
        """
        #: MicroscopeModel: The microscope model associated with the loop control.
        self.model = model

        #: bool: A boolean value indicating whether to step by frame or by step.
        self.step_by_frame = step_by_frame

        #: int: The remaining number of steps or frames.
        self.steps = steps

        #: bool: Flags indicate if nested loops
        self.is_nested = is_nested

        #: int: The remaining number of steps.
        self.signals = 1

        #: int: The remaining number of frames.
        self.data_frames = 1

        #: bool: Initialization flag
        self.initialized = VariableWithLock(bool)
        self.initialized.value = False

        #: int: signal/data end num
        self.end_count = VariableWithLock(int)
        self.end_count.value = 0

        #: dict: A dictionary defining the configuration for the loop control process.
        self.config_table = {
            "signal": {
                "init": self.pre_func,
                "main": self.signal_func,
            },
            "data": {"init": self.pre_func, "main": self.data_func},
        }

    def pre_func(self):
        """Initialize loop parameters"""
        if self.initialized.value:
            return

        with self.initialized as initialized:
            if initialized.value:
                return

            steps = self.get_steps()

            self.signals = steps
            self.data_frames = steps
            initialized.value = True

            logger.debug(f"LoopByCount-initialize: {self.signals}, {self.data_frames}")

    def signal_func(self):
        """Control the signal acquisition loop and update the remaining steps.

        This method controls the signal acquisition loop by decrementing the remaining
        steps. It determines whether to continue the loop or exit based on the
        remaining count.

        Returns:
        -------
        bool
            A boolean value indicating whether to continue the loop.
        """
        self.signals -= 1
        if self.signals <= 0:
            if self.is_nested:
                self.synchronize("signal")
            return False
        return True

    def data_func(self, frame_ids):
        """Control the data acquisition loop and update the remaining frames or steps.

        This method controls the data acquisition loop by decrementing the remaining
        frames or steps. It determines whether to continue the loop or exit based on
        the remaining count.

        Parameters:
        ----------
        frame_ids : list
            A list of frame IDs for which data acquisition should be performed.

        Returns:
        -------
        bool
            A boolean value indicating whether to continue the loop.
        """
        if self.step_by_frame:
            self.data_frames -= len(frame_ids)
        else:
            self.data_frames -= 1
        if self.data_frames <= 0:
            if self.is_nested:
                self.synchronize("data")
            return False
        return True

    def get_steps(self):
        """Get number of steps

        Returns:
        --------
        int
            Number of steps.
        """
        steps = self.steps

        if type(steps) is int:
            return steps
        if steps == "channels":
            return len(self.model.active_microscope.available_channels)
        elif steps == "positions":
            return len(self.model.configuration["multi_positions"]) - 1
        else:
            try:
                parameters = steps.split(".")
                config_ref = reduce((lambda pre, n: f"{pre}['{n}']"), parameters, "")
                steps = eval(f"self.model.configuration{config_ref}")
            except:  # noqa
                return 1

            if type(steps) in [list, ListProxy]:
                return len(steps)
            else:
                return int(steps)

    def synchronize(self, thread_name):
        """Synchronize signal and data function

        Parameters:
        ----------
        thread_name : bool
            Signal or Data
        """
        logger.debug(f"LoopByCount-Synchronize {thread_name}")
        with self.end_count as end_count:
            if end_count.value == 1:
                self.initialized.value = False
                end_count.value += 1
                return
            end_count.value += 1

        while self.end_count.value < 2:
            time.sleep(0.001)

        self.end_count.value = 0

        logger.debug(f"LoopByCount-Synchronize {thread_name} ends.")


class PrepareNextChannel:
    """PrepareNextChannel class for preparing microscopes for the next imaging channel.

    This class provides functionality to prepare multiple microscopes, including virtual
    microscopes and the primary microscope, for the next imaging channel during
    microscopy experiments.

    Notes:
    ------
    - This class is used to prepare multiple microscopes for the next imaging
    channel, ensuring that both virtual microscopes and the primary microscope are
    ready for the next step in microscopy experiments.

    - The `PrepareNextChannel` class is typically used to manage the preparation of
    microscopes before transitioning to a new imaging channel.

    - The channel preparation process involves calling the `prepare_next_channel()`
    method for each virtual microscope and the active microscope.

    - The `config_table` attribute is used to define the configuration for the
    channel preparation process, specifically the main preparation function.
    """

    def __init__(self, model):
        """Initialize the PrepareNextChannel class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for channel preparation.
        """
        #: MicroscopeModel: Microscope model associated with the channel preparation.
        self.model = model

        #: dict: A dictionary defining the configuration for the channel preparation
        self.config_table = {"signal": {"main": self.signal_func}}

    def signal_func(self):
        """Prepare virtual and active microscopes for the next imaging channel.

        This method prepares virtual microscopes, if any, followed by the active
        microscope for the next imaging channel.

        Returns:
        -------
        bool
            A boolean value indicating the success of the channel preparation process.
        """
        # prepare virtual microscopes before the primary microscope
        for microscope_name in self.model.virtual_microscopes:
            self.model.virtual_microscopes[microscope_name].prepare_next_channel()

        self.model.active_microscope.prepare_next_channel()

        return True


class MoveToNextPositionInMultiPositionTable:
    """MoveToNextPositionInMultiPositionTable class for advancing in a multi-position
    table.

    This class provides functionality to move to the next position in a multi-position
    table and control the data thread accordingly.

    Notes:
    ------
    - This class is used to advance to the next position in a multi-position table,
    controlling the data thread based on stage distance thresholds.

    - The `MoveToNextPositionInMultiPositionTable` class is typically used to automate
     position changes during microscopy experiments, ensuring proper data thread
     management.

    - The position control process involves moving to the next position in the table,
    pausing the data thread if necessary, and resuming it after the movement.

    - The `config_table` attribute defines the configuration for the position control
    process, including signal acquisition and cleanup steps.
    """

    def __init__(self, model, resolution_value=None, zoom_value=None, offset=None):
        """Initialize the MoveToNextPositionInMultiPositionTable class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for position control.
        resolution_value : str
            The resolution/microscope name of the current multiposition table
        zoom_value : str
            The zoom name. For example "1x", "2x", ...
        offset : list
            The position offset [x, y, z, theta, f]
        """
        #: MicroscopeModel: The microscope model associated with position control.
        self.model = model

        #: dict: A dictionary defining the configuration for the position control
        self.config_table = {
            "signal": {
                "init": self.pre_signal_func,
                "main": self.signal_func,
                "cleanup": self.cleanup,
            },
            "node": {"device_related": True},
        }

        #: int: The current index of the position being acquired in the multi-position
        self.pre_z = None

        #: int: The current index of the position being acquired in the multi-position
        self.current_idx = 0

        #: dict: A dictionary defining the configuration for the position control
        self.multiposition_table = []

        #: int: The total number of positions in the multi-position table.
        self.position_count = 0

        #: int: The stage distance threshold for pausing the data thread.
        self.stage_distance_threshold = 1000

        #: str: The microscope name/resolution name
        self.resolution_value = resolution_value

        #: str: The zoom value
        self.zoom_value = zoom_value

        #: dict: stage offset table
        self.offset = offset

        #: bool: The flag indicates whether this node is initialized
        self.initialized = False

    def pre_signal_func(self):
        """Calculate stage offset if applicable."""
        if self.initialized:
            return
        self.initialized = True

        # the first row will be headers
        self.multiposition_table = self.model.configuration["multi_positions"][1:]
        headers = self.model.configuration["multi_positions"][0]

        #: list: The list of stage axes (e.g., X, Y, Z, THETA, F)
        self.stage_axes = list(self.model.active_microscope.stages.keys())

        #: list: The list of axes indices corresponding to the stage axes in the
        #: multi-position table.
        self.axes_index = [headers.index(axis.upper()) for axis in self.stage_axes]
        self.position_count = len(self.multiposition_table)
        axes_num = len(self.stage_axes)
        if type(self.offset) is str:
            try:
                self.offset = ast.literal_eval(self.offset)
            except SyntaxError:
                self.offset = [0] * axes_num
        if not self.offset or type(self.offset) is not list:
            self.offset = [0] * axes_num

        # assert offset has values for each axis
        if len(self.offset) < axes_num:
            self.offset[len(self.offset) : axes_num] = [0] * (
                axes_num - len(self.offset)
            )
        for i in range(axes_num):
            try:
                self.offset[i] = float(self.offset[i])
            except (ValueError, TypeError):
                self.offset[i] = 0

        curr_resolution = self.model.active_microscope_name
        curr_zoom = self.model.active_microscope.zoom.zoomvalue
        if not self.resolution_value or not self.zoom_value:
            return
        if curr_resolution == self.resolution_value and curr_zoom == self.zoom_value:
            return
        # calculate offset
        if curr_resolution != self.resolution_value:
            stage_offset = self.model.configuration["configuration"]["microscopes"][
                self.resolution_value
            ]["stage"]
            curr_stage_offset = self.model.configuration["configuration"][
                "microscopes"
            ][curr_resolution]["stage"]
            for i, axis in enumerate(self.stage_axes):
                self.offset[i] = (
                    self.offset[i]
                    + curr_stage_offset.get(axis + "_offset", 0)
                    - stage_offset.get(axis + "_offset", 0)
                )
        else:
            solvent = self.model.configuration["experiment"]["Saving"]["solvent"]
            stage_solvent_offsets = self.model.active_microscope.zoom.stage_offsets
            if solvent in stage_solvent_offsets.keys():
                stage_offset = stage_solvent_offsets[solvent]
                for i, axis in enumerate(self.stage_axes):
                    if axis not in stage_offset.keys():
                        continue
                    try:
                        self.offset[i] = self.offset[i] + float(
                            stage_offset[axis][self.zoom_value][curr_zoom]
                        )
                    except (ValueError, KeyError):
                        print(
                            f"*** Offsets from {self.zoom_value} to {curr_zoom} are "
                            f"not implemented! There is not enough information in the "
                            f"configuration.yaml file!"
                        )
        logger.debug(f"Using stage offset {self.offset}")

    def signal_func(self):
        """Move to the next position in the multi-position table and control the data
        thread.

        This method advances to the next position in the multi-position table,
        controls the data thread based on stage distance thresholds, and updates
        position-related information.

        Returns:
        -------
        bool
            A boolean value indicating whether to continue the position control process.
        """
        logger.debug(
            f"multi-position current idx: {self.current_idx}, {self.position_count}"
        )
        if self.current_idx >= self.position_count:
            return False
        # add offset
        pos_dict = dict(
            zip(
                self.stage_axes,
                [
                    self.multiposition_table[self.current_idx][self.axes_index[i]]
                    + self.offset[i]
                    for i in range(len(self.axes_index))
                ],
            )
        )
        # pause data thread if necessary
        if self.current_idx == 0:
            temp = self.model.get_stage_position()
            pre_stage_pos = dict(
                map(
                    lambda k: (k, temp[f"{k}_pos"]),
                    self.stage_axes,
                )
            )
        else:
            pre_stage_pos = dict(
                zip(
                    self.stage_axes,
                    [
                        self.multiposition_table[self.current_idx - 1][
                            self.axes_index[i]
                        ]
                        + self.offset[i]
                        for i in range(len(self.axes_index))
                    ],
                )
            )
        delta_distances = [
            abs(pos_dict[axis] - pre_stage_pos[axis]) for axis in self.stage_axes
        ]
        should_pause_data_thread = any(
            distance > self.stage_distance_threshold for distance in delta_distances
        )
        if should_pause_data_thread:
            self.model.pause_data_thread()

        self.current_idx += 1
        # Make sure to go back to the beginning if using LoopByCount
        if self.current_idx == self.position_count:
            self.current_idx = 0

        abs_pos_dict = dict(map(lambda k: (f"{k}_abs", pos_dict[k]), pos_dict.keys()))
        logger.debug(f"MoveToNextPositionInMultiPosition: " f"{pos_dict}")
        self.model.move_stage(abs_pos_dict, wait_until_done=True)

        logger.debug("MoveToNextPositionInMultiPosition: move done")

        # resume data thread
        if should_pause_data_thread:
            self.model.resume_data_thread()
        self.model.active_microscope.central_focus = None
        if self.pre_z != pos_dict["z"]:
            self.pre_z = pos_dict["z"]
            return True

    def cleanup(self):
        """Cleanup method to resume the data thread.

        This method is responsible for resuming the data thread after position control.
        """
        self.model.resume_data_thread()


class StackPause:
    """StackPause class for pausing stack acquisition.

    This class provides functionality to pause stack acquisition for a specified
    number of timepoints or based on a defined pause time. It manages the data thread
    accordingly.

    Notes:
    ------
    - This class is used to pause stack acquisition for a specified number of timepoints
    or based on a defined pause time during microscopy experiments.

    - The `StackPause` class allows for flexible control of stack acquisition pauses,
    ensuring synchronization with data acquisition.

    - The stack pause control process involves managing the data thread, calculating
    pause times, and handling stack acquisition pauses.

    - The `config_table` attribute defines the configuration for the stack pause control
    process, specifically the main pause function.
    """

    def __init__(self, model, pause_num="experiment.MicroscopeState.timepoints"):
        """Initialize the StackPause class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for stack acquisition control.

        pause_num : int or str, optional
            The number of timepoints to pause stack acquisition or a configuration
            reference to determine the pause count dynamically. Default is
            "experiment.MicroscopeState.timepoints".
        """
        self.model = model
        self.pause_num = pause_num
        if type(pause_num) is str:
            try:
                parameters = pause_num.split(".")
                config_ref = reduce((lambda pre, n: f"{pre}['{n}']"), parameters, "")
                exec(f"self.pause_num = int(self.model.configuration{config_ref})")
            except:  # noqa
                self.pause_num = 1
        self.config_table = {"signal": {"main": self.signal_func}}

    def signal_func(self):
        """Pause stack acquisition based on timepoints or pause time and manage the
        data thread.

        This method pauses stack acquisition based on the remaining timepoints or
        defined pause time. It manages the data thread accordingly during the pause.
        """
        self.pause_num -= 1
        if self.pause_num <= 0:
            return
        pause_time = float(
            self.model.configuration["experiment"]["MicroscopeState"]["stack_pause"]
        )
        if pause_time <= 0:
            return
        current_channel = f"channel_{self.model.active_microscope.current_channel}"
        current_exposure_time = (
            float(
                self.model.configuration["experiment"]["MicroscopeState"]["channels"][
                    current_channel
                ]["camera_exposure_time"]
            )
            / 1000.0
        )
        if pause_time < 5 * current_exposure_time:
            time.sleep(pause_time)
        else:
            self.model.pause_data_thread()
            pause_time -= 2 * current_exposure_time
            while pause_time > 0:
                pt = min(pause_time, 0.1)
                time.sleep(pt)
                if self.model.stop_acquisition:
                    self.model.resume_data_thread()
                    return
                pause_time -= 0.1
            self.model.resume_data_thread()


class ZStackAcquisition:
    """ZStackAcquisition class for controlling z-stack acquisition in microscopy.

    This class provides functionality to control z-stack acquisition, including managing
    z and focus positions, acquiring image data, and handling multi-channel
    acquisitions.

    Notes:
    ------
    - This class is used to control z-stack acquisition during microscopy experiments,
    allowing for position cycling, multi-channel acquisitions, and image saving.

    - The z-stack acquisition process involves initializing parameters, controlling
    position and focus movements, handling data acquisition, and managing data threads.

    - The `config_table` attribute defines the configuration for the z-stack acquisition
    process, including signal acquisition, data handling, and node type.
    """

    def __init__(
        self,
        model,
        get_origin=False,
        saving_flag=False,
        saving_dir="z-stack",
        force_multiposition=False,
    ):
        """Initialize the ZStackAcquisition class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for z-stack acquisition control.
        get_origin : bool, optional
            Flag to determine whether to get the z and focus origin positions.
            Default is False.
        saving_flag : bool, optional
            Flag to enable image saving during z-stack acquisition. Default is False.
        saving_dir : str, optional
            The subdirectory for saving z-stack images. The default is "z-stack".

        """
        #: MicroscopeModel: The microscope model associated with z-stack acquisition.
        self.model = model

        #: int: The current channel being acquired in the z-stack
        self.current_channel_in_list = None

        #: bool: Flag to determine whether to get the z and focus origin positions.
        self.get_origin = get_origin

        #: int: The number of z steps in the z-stack.
        self.number_z_steps = 0

        #: float: The start z position for the z-stack.
        self.start_z_position = 0

        #: float: The z stack distance for the z-stack.
        self.z_stack_distance = None

        #: float: The focus stack distance for the z-stack.
        self.f_stack_distance = None

        #: float: The z position of the channel being acquired in the z-stack
        self.restore_z = None

        #: float: The f position of the channel being acquired in the z-stack
        self.restore_f = None

        #: float: The start focus position for the z-stack.
        self.start_focus = 0

        #: float: The z step size for the z-stack.
        self.z_step_size = 0

        #: float: The z stack distance for the z-stack.
        self.focus_step_size = 0

        #: dict: A dictionary defining the multi-position table for z-stack acquisition.
        self.positions = {}

        #: list: The list of stage axes (e.g., X, Y, Z, THETA, F)
        self.position_headers = []

        #: list: A list of each header position in the multi-position table.
        self.axes_index = []

        #: dict: A dictionary defining the current position in the multi-position table.
        self.current_position = None

        #: int: The current index of the position being acquired in the multi-position
        self.current_position_idx = 0

        #: int: The z position of the channel being acquired in the z-stack
        self.current_z_position = 0

        #: int: The f position of the channel being acquired in the z-stack
        self.current_focus_position = 0

        #: bool: Flag to determine whether to move to a new position
        self.need_to_move_new_position = True

        #: bool: Flag to determine whether to move the z position
        self.need_to_move_z_position = True

        #: bool: Flag to determine whether to pause the data thread.
        self.should_pause_data_thread = None

        # TODO: distance > 1000 should not be hardcoded and somehow related to
        #  different kinds of stage devices.
        #: int: The stage distance threshold for pausing the data thread.
        self.stage_distance_threshold = 1000

        #: dict: A dictionary of the previous position in the multi-position table.
        self.pre_position = None

        #: int: The number of times the z position has been moved
        self.z_position_moved_time = 0

        #: dict: A dictionary defining the defocus values between channels
        self.defocus = None

        #: str: The stack cycling mode for z-stack acquisition.
        self.stack_cycling_mode = "per_stack"

        #: int: The number of channels in the z-stack.
        self.channels = 1

        #: bool: Force multiposition
        self.force_multiposition = force_multiposition

        #: int: The number of frames received by the data thread.
        self.received_frames = None

        # TODO: This does not include timepoints.
        #: int: The number of frames anticipated by the data thread.
        self.total_frames = None

        #: int: The number of timepoints in the z-stack acquisition.
        self.timepoints = None

        #: ImageWriter: An image writer object for saving z-stack images.
        self.image_writer = None
        if saving_flag:
            self.image_writer = ImageWriter(model, sub_dir=saving_dir)

        self.prepare_next_channel = PrepareNextChannel(model)

        #: dict: A dictionary defining the configuration for the z-stack acquisition
        self.config_table = {
            "signal": {
                "init": self.pre_signal_func,
                "main": self.signal_func,
                "end": self.signal_end,
            },
            "data": {
                "init": self.pre_data_func,
                "main": self.in_data_func,
                "end": self.end_data_func,
                "cleanup": self.cleanup_data_func,
            },
            "node": {"node_type": "multi-step", "device_related": True},
        }

    def get_microscope_state(self, microscope_state: dict) -> None:
        """Get the microscope state from the configuration.

        Parameters:
        ----------
        microscope_state : dict
            The microscope state configuration dictionary.
        """
        self.timepoints = microscope_state["timepoints"]
        self.stack_cycling_mode = microscope_state["stack_cycling_mode"]

        self.stage_axes = list(self.model.active_microscope.stages.keys())
        # primary z and f axes, secondary stack settings
        self.primary_z_axis = microscope_state.get("primary_z_axis", "z")
        self.primary_f_axis = microscope_state.get("primary_f_axis", "f")
        self.secondary_stack_settings = microscope_state.get("secondary_stack_settings", {})
        self.tiling_axes = list(
            set(self.stage_axes) - 
            set([self.primary_z_axis, self.primary_f_axis]) - 
            set(self.secondary_stack_settings.keys())
        )
        

        # get available channels
        self.channels = len(
            list(
                filter(
                    lambda channel: channel["is_selected"],
                    microscope_state["channels"].values(),
                )
            )
        )
        self.current_channel_in_list = 0

    def get_z_stack_parameters(self, microscope_state: dict) -> None:
        """Get z-stack parameters from the configuration.

        Parameters:
        ----------
        microscope_state : dict
            The microscope state configuration dictionary.
        """
        self.number_z_steps = int(microscope_state["number_z_steps"])
        self.start_z_position = float(microscope_state["start_position"])
        self.z_step_size = float(microscope_state["step_size"])
        self.z_stack_distance = abs(
            self.start_z_position - float(microscope_state["end_position"])
        )

    def get_f_stack_parameters(self, microscope_state: dict) -> None:
        """Get focus stack parameters from the configuration.

        Parameters:
        ----------
        microscope_state : dict
            The microscope state configuration dictionary.
        """
        self.start_focus = float(microscope_state["start_focus"])
        end_focus = float(microscope_state["end_focus"])
        self.focus_step_size = (end_focus - self.start_focus) / self.number_z_steps
        self.f_stack_distance = abs(end_focus - self.start_focus)

    def pre_signal_func(self) -> None:
        """Initialize z-stack acquisition parameters before the signal stage.

        This method initializes z-stack acquisition parameters, including position,
        focus, and data thread management, before the signal stage.
        """
        microscope_state = self.model.configuration["experiment"]["MicroscopeState"]
        self.get_microscope_state(microscope_state)
        self.get_z_stack_parameters(microscope_state)
        self.get_f_stack_parameters(microscope_state)

        # restore z, f
        pos_dict = self.model.get_stage_position()
        self.restore_z = pos_dict["z_pos"]
        self.restore_f = pos_dict["f_pos"]

        # position: x, y, z, theta, f
        if bool(microscope_state["is_multiposition"]) or self.force_multiposition:
            self.position_headers = self.model.configuration["multi_positions"][0]
            self.positions = self.model.configuration["multi_positions"][1:]
        else:
            self.position_headers = [axis.upper() for axis in self.stage_axes]
            self.positions = [[float(pos_dict[f"{axis}_pos"]) for axis in self.stage_axes]]

            # get origin for primary z and f
            if not self.get_origin:
                self.positions[0][self.stage_axes.index(self.primary_z_axis)] = microscope_state.get(
                    "stack_z_origin",
                    pos_dict[f"{self.primary_z_axis}_pos"]
                )
                self.positions[0][self.stage_axes.index(self.primary_f_axis)] = microscope_state.get(
                    "stack_focus_origin",
                    pos_dict[f"{self.primary_f_axis}_pos"]
                )

        self.axes_index = [self.position_headers.index(axis.upper()) for axis in self.stage_axes]

        # Set up the next channel down here to ensure defocus isn't merged into
        # restore f_pos, positions
        self.model.active_microscope.central_focus = None
        self.model.active_microscope.current_channel = 0
        for microscope_name in self.model.virtual_microscopes:
            self.model.virtual_microscopes[microscope_name].current_channel = 0
        self.prepare_next_channel.signal_func()

        logger.info(
            f"ZStackAcquisition: Positions {self.positions}, "
            f"Timepoints {self.timepoints}, "
            f"Starting Focus {self.start_focus}, "
            f"Starting Z-Position {self.start_z_position}"
        )
        self.current_position_idx = 0
        self.current_position = dict(
            zip(self.stage_axes, [self.positions[0][i] for i in self.axes_index])
        )
        self.z_position_moved_time = 0
        self.need_to_move_new_position = True
        self.need_to_move_z_position = True
        self.should_pause_data_thread = False

        self.defocus = [
            v["defocus"]
            for v in microscope_state["channels"].values()
            if v["is_selected"]
        ]

    def signal_func(self):
        """Control z-stack acquisition, move positions, and manage data threads.

        This method controls the z-stack acquisition process, including moving positions
        and focus, managing data threads, and handling data acquisition during the
        signal stage.

        Returns:
        -------
        bool
            A boolean value indicating whether to continue the z-stack acquisition
            process.
        """
        if self.model.stop_acquisition:
            return False
        data_thread_is_paused = False

        # move stage X, Y, Theta
        if self.need_to_move_new_position:
            self.need_to_move_new_position = False
            self.pre_position = self.current_position
            self.current_position = dict(
                zip(
                    self.stage_axes,
                    [self.positions[self.current_position_idx][i] for i in self.axes_index],
                )
            )

            # calculate first z, f position
            self.current_z_position = self.start_z_position + self.current_position[self.primary_z_axis]
            self.current_focus_position = self.start_focus + self.current_position[self.primary_f_axis]
            if self.defocus is not None:
                self.current_focus_position += self.defocus[
                    self.current_channel_in_list
                ]

            pos_dict = dict(
                map(
                    lambda ax: (
                        f"{ax}_abs",
                        self.current_position[ax],
                    ),
                    self.tiling_axes,
                )
            )

            if self.current_position_idx > 0:
                delta_distances = [self.current_position[axis] - self.pre_position[axis] for axis in self.tiling_axes if axis != "theta"]
                delta_distances.append(
                    self.current_position[self.primary_z_axis]
                    - self.pre_position[self.primary_z_axis]
                    + self.z_stack_distance
                )
                delta_distances.append(
                    self.current_position[self.primary_f_axis]
                    - self.pre_position[self.primary_f_axis]
                    + self.f_stack_distance
                )
            else:
                axes_num = len(self.tiling_axes) + 2 - (1 if "theta" in self.tiling_axes else 0)
                delta_distances = [0] * axes_num

            # displacement = [delta_z, delta_f, delta_x, delta_y]
            # Check the distance between the current position and previous position,
            # if it is too far, then we can call self.model.pause_data_thread() and
            # self.model.resume_data_thread() after the stage has completed the move
            # to the next position.

            self.should_pause_data_thread = any(
                distance > self.stage_distance_threshold
                for distance in delta_distances
            )
            if self.should_pause_data_thread:
                self.model.pause_data_thread()
                data_thread_is_paused = True

            self.model.move_stage(pos_dict, wait_until_done=True)

        # Potentially pause the data thread and move z, f position
        if self.need_to_move_z_position:
            if self.should_pause_data_thread and not data_thread_is_paused:
                self.model.pause_data_thread()
                logger.info("Data thread paused.")

            stack_pos = [
                (f"{self.primary_z_axis}_abs", self.current_z_position),
                (f"{self.primary_f_axis}_abs", self.current_focus_position)
            ]
            for axis, offset in self.secondary_stack_settings.items():
                stack_pos.append((f"{axis}_abs", self.current_z_position + offset))
            self.model.move_stage(
                dict(stack_pos),
                wait_until_done=True,
            )

        if self.should_pause_data_thread:
            self.model.resume_data_thread()
            self.should_pause_data_thread = False

        self.model.mark_saving_flags([self.model.frame_id])

        return True

    def signal_end(self) -> bool:
        """Handle the end of the signal stage and position cycling.

        This method handles the end of the signal stage, including position cycling and
        channel updates for multichannel acquisitions.

        Returns:
        -------
        bool
            A boolean value indicating whether to end the current node.
        """

        # end this node
        if self.model.stop_acquisition:
            return True

        if self.stack_cycling_mode != "per_stack":
            # update the channel for each z position in 'per_slice'
            if self.defocus is not None:
                self.current_focus_position -= self.defocus[
                    self.current_channel_in_list
                ]
            self.update_channel()
            self.need_to_move_z_position = self.current_channel_in_list == 0

        # in 'per_slice', move to the next z position if all the channels have been
        # acquired
        if self.need_to_move_z_position:
            # next z, f position
            self.current_z_position += self.z_step_size
            self.current_focus_position += self.focus_step_size

            # update z position moved time
            self.z_position_moved_time += 1

        # decide whether to move X, Y, Theta
        if self.z_position_moved_time >= self.number_z_steps:
            self.z_position_moved_time = 0
            # calculate first z, f position
            self.current_z_position = self.start_z_position + self.current_position[self.primary_z_axis]
            self.current_focus_position = self.start_focus + self.current_position[self.primary_f_axis]
            if (
                self.z_stack_distance > self.stage_distance_threshold
                or self.f_stack_distance > self.stage_distance_threshold
            ):
                self.should_pause_data_thread = True

            # after running through a z-stack, update channel
            if self.stack_cycling_mode == "per_stack":
                self.update_channel()
                # if run through all the channels, move to the next position
                if self.current_channel_in_list == 0:
                    self.need_to_move_new_position = True
            else:
                self.need_to_move_new_position = True

            if self.need_to_move_new_position:
                # move to the next position
                self.current_position_idx += 1

        if self.current_position_idx >= len(self.positions):
            self.current_position_idx = 0
            # restore z, f, and secondary z if any
            stack_pos = [
                (f"{self.primary_z_axis}_abs", self.restore_z),
                (f"{self.primary_f_axis}_abs", self.restore_f)
            ]
            for axis, offset in self.secondary_stack_settings.items():
                stack_pos.append((f"{axis}_abs", self.restore_z + offset))
            self.model.move_stage(
                dict(stack_pos),
                wait_until_done=False,
            )  # Update position
            return True

        return False

    def update_channel(self) -> None:
        """Update the active channel during multichannel acquisition.

        This method updates the active channel for multichannel acquisitions, allowing
        cycling through channels.
        """
        self.current_channel_in_list = (
            self.current_channel_in_list + 1
        ) % self.channels
        # not update DAQ tasks if there is a NI Galvo stage
        self.prepare_next_channel.signal_func()
        if self.defocus is not None:
            self.current_focus_position += self.defocus[self.current_channel_in_list]

    def pre_data_func(self) -> None:
        """Initialize data-related parameters before data acquisition.

        This method initializes data-related parameters before data acquisition,
        including the count of received and expected frames.
        """

        self.received_frames = 0
        self.total_frames = self.channels * self.number_z_steps * len(self.positions)

    def in_data_func(self, frame_ids: list) -> None:
        """Handle incoming data frames during data acquisition.

        This method handles incoming data frames during data acquisition, updating the
        count of received frames and saving images if enabled.

        Parameters:
        ----------
        frame_ids : list
            A list of frame IDs received during data acquisition.

        """
        self.received_frames += len(frame_ids)
        if self.image_writer is not None:
            self.image_writer.save_image(frame_ids)

    def end_data_func(self) -> bool:
        """Check if all expected data frames have been received.

        This method checks whether all expected data frames have been received during
        data acquisition.

        Returns:
        -------
        bool
            A boolean value indicating whether all expected data frames have been
            received.
        """

        return self.received_frames >= self.total_frames

    def cleanup_data_func(self) -> None:
        """Perform cleanup actions after data acquisition if image saving is enabled.

        This method performs cleanup actions after data acquisition, such as cleaning up
         image writing, if image saving is enabled.
        """
        if self.image_writer:
            self.image_writer.cleanup()


class FindTissueSimple2D:
    """FindTissueSimple2D class for detecting tissue and gridding out the imaging
    space in  2D.

    This class is responsible for detecting tissue, thresholding, and gridding out the
    space for 2D imaging. It processes acquired frames to determine regions of
    interest (tissue), calculates offsets, and generates grid positions for imaging.
    """

    def __init__(
        self,
        model,
        overlap=0.1,
        target_resolution="Nanoscale",
        target_zoom="N/A",
    ):
        """Initialize the FindTissueSimple2D class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for tissue detection and gridding.
        overlap : float, optional
            The overlap percentage between grid tiles. Default is 0.1 (10%).
        target_resolution : str, optional
            The target resolution for imaging (e.g., "Nanoscale"). Default is
            "Nanoscale".
        target_zoom : str, optional
            The target zoom level for imaging. Default is "N/A".
        """
        #: MicroscopeModel: The microscope model associated with tissue detection and
        self.model = model

        #: dict: A dictionary defining the configuration for tissue detection and
        self.config_table = {"signal": {}, "data": {"main": self.data_func}}

        #: float: The overlap percentage between grid tiles.
        self.overlap = overlap

        #: str: The target resolution for imaging (e.g., "Nanoscale").
        self.target_resolution = target_resolution

        #: str: The target zoom level for imaging.
        self.target_zoom = target_zoom

    def data_func(self, frame_ids):
        """Process acquired frames for tissue detection, thresholding, and grid
        calculation.

        This method processes acquired frames to detect tissue regions, apply
        thresholding, and calculate grid positions for imaging. It performs the
        following steps:
        - Downsamples the acquired image based on desired magnification change.
        - Applies thresholding using Otsu's method.
        - Calculates the bounding box of the tissue region.
        - Computes offsets for gridding based on acquired and target resolutions.
        - Grids out the 2D imaging space based on the specified overlap percentage.

        Parameters:
        ----------
        frame_ids : list
            A list of frame IDs corresponding to acquired frames.
        """

        from skimage import filters
        from skimage.transform import downscale_local_mean
        import numpy as np
        from navigate.tools.multipos_table_tools import (
            compute_tiles_from_bounding_box,
            calc_num_tiles,
        )

        for idx in frame_ids:
            img = self.model.data_buffer[idx]

            # Get current mag
            microscope_name = self.model.configuration["experiment"]["MicroscopeState"][
                "microscope_name"
            ]
            zoom = self.model.configuration["experiment"]["MicroscopeState"]["zoom"]
            curr_pixel_size = self.model.configuration["configuration"]["microscopes"][
                microscope_name
            ]["zoom"][zoom]["pixel_size"]
            # get target pixel size
            pixel_size = self.model.configuration["configuration"]["microscopes"][
                self.target_resolution
            ]["zoom"][self.target_zoom]["pixel_size"]

            # Downsample according to the desired magnification change. Note, we
            # could downsample by whatever we want.
            ds = int(curr_pixel_size / pixel_size)
            ds_img = downscale_local_mean(img, (ds, ds))

            # Threshold
            thresh_img = ds_img > filters.threshold_otsu(img)

            # Find the bounding box
            # In the real-deal, non-transposed image, x increase corresponds to a
            # decrease in row number y increase responds to an increase in row number
            # This means the smallest x coordinate is actually the largest
            x, y = np.where(thresh_img)
            # + 0.5 accounts for center of FOV
            x_start, x_end = -curr_pixel_size * ds * (
                np.max(x) + 0.5
            ), -curr_pixel_size * ds * (np.min(x) + 0.5)
            y_start, y_end = curr_pixel_size * ds * (
                np.min(y) + 0.5
            ), curr_pixel_size * ds * (np.max(y) + 0.5)
            xd, yd = abs(x_start - x_end), y_end - y_start

            # grab z, theta, f starting positions
            z_start = self.model.configuration["experiment"]["StageParameters"]["z"]
            r_start = self.model.configuration["experiment"]["StageParameters"]["theta"]
            if self.target_resolution == "Nanoscale":
                f_start = 0  # very different range of focus values in high-res
            else:
                f_start = self.model.configuration["experiment"]["StageParameters"]["f"]

            # Update x and y start to initialize from the upper-left corner of the
            # current image, since this is how np.where indexed them. The + 0.5 in
            # x_start/y_start calculation shifts their starts back to the center of the
            # field of view.
            curr_fov_x = (
                float(
                    self.model.configuration["experiment"]["CameraParameters"][
                        microscope_name
                    ]["x_pixels"]
                )
                * curr_pixel_size
            )
            curr_fov_y = (
                float(
                    self.model.configuration["experiment"]["CameraParameters"][
                        microscope_name
                    ]["y_pixels"]
                )
                * curr_pixel_size
            )
            x_start += (
                self.model.configuration["experiment"]["StageParameters"]["x"]
                + curr_fov_x / 2
            )
            y_start += (
                self.model.configuration["experiment"]["StageParameters"]["y"]
                - curr_fov_y / 2
            )

            # stage offset
            x_start += float(
                self.model.configuration["configuration"]["microscopes"][
                    self.target_resolution
                ]["stage"]["x_offset"]
            ) - float(
                self.model.configuration["configuration"]["microscopes"][
                    microscope_name
                ]["stage"]["x_offset"]
            )
            y_start += float(
                self.model.configuration["configuration"]["microscopes"][
                    self.target_resolution
                ]["stage"]["y_offset"]
            ) - float(
                self.model.configuration["configuration"]["microscopes"][
                    microscope_name
                ]["stage"]["y_offset"]
            )
            z_start += float(
                self.model.configuration["configuration"]["microscopes"][
                    self.target_resolution
                ]["stage"]["z_offset"]
            ) - float(
                self.model.configuration["configuration"]["microscopes"][
                    microscope_name
                ]["stage"]["z_offset"]
            )
            r_start += float(
                self.model.configuration["configuration"]["microscopes"][
                    self.target_resolution
                ]["stage"]["r_offset"]
            ) - float(
                self.model.configuration["configuration"]["microscopes"][
                    microscope_name
                ]["stage"]["r_offset"]
            )

            # grid out the 2D space
            fov_x = (
                float(
                    self.model.configuration["experiment"]["CameraParameters"][
                        microscope_name
                    ]["x_pixels"]
                )
                * pixel_size
            )
            fov_y = (
                float(
                    self.model.configuration["experiment"]["CameraParameters"][
                        microscope_name
                    ]["y_pixels"]
                )
                * pixel_size
            )
            x_tiles = calc_num_tiles(xd, self.overlap, fov_x)
            y_tiles = calc_num_tiles(yd, self.overlap, fov_y)

            table_values = compute_tiles_from_bounding_box(
                x_start,
                x_tiles,
                fov_x,
                self.overlap,
                y_start,
                y_tiles,
                fov_y,
                self.overlap,
                z_start,
                1,
                0,
                self.overlap,
                r_start,
                1,
                0,
                self.overlap,
                f_start,
                1,
                0,
                self.overlap,
            )

            self.model.event_queue.put(("multiposition", table_values))



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import time
import serial
import threading
import logging
from typing import Optional, Tuple, Union

# Third-party imports
import numpy as np

# Local application imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class MP285:
    """Sutter MP285 3D Stage API

    The USB interface expects a baud rate of 9600, 8 data bits, 1 stop bits,
    no parity, and no flow control.

    The serial interface expects a baud rate of 9600, 8 data bits, 1 stop bits,
    no parity, and hardware or RTS/CTS flow control.

    Little-endian byte order is used for all commands and responses.

    Each command sequence consists of at least one byte, the first of which is the
    “command byte”. Those commands that have parameters or arguments require a
    sequence of bytes that follow the command byte. No delimiters are used between
    command sequence arguments, and command sequence terminators are used in most cases.

    Most command sequences have a terminator: ASCII CR (Carriage Return; 13 decimal,
    0D hexadecimal) to indicate that the task associated with the command has completed.
    When the controller completes the task associated with a command, it sends ASCII CR
    back to the host computer indicating that it is ready to receive a new command.
    If a command returns data, the last byte returned is the task-completed indicator.
    """

    def __init__(self, com_port: str, baud_rate: int, timeout=0.25) -> None:
        """Initialize the MP-285 stage.

        Parameters
        ----------
        com_port : str
            COM port of the MP-285 stage.
        baud_rate : int
            Baud rate of the MP-285 stage.
        timeout : float
            Timeout for the serial connection.
        """

        #: serial.Serial: Serial connection to the MP-285 stage
        self.serial = serial.Serial()
        self.serial.port = com_port
        self.serial.baudrate = baud_rate
        self.serial.timeout = timeout
        self.serial.parity = serial.PARITY_NONE
        self.serial.bytesize = serial.EIGHTBITS
        self.serial.stopbits = serial.STOPBITS_ONE
        self.serial.xonxoff = False
        self.serial.rtscts = True

        #: int: Speed of the stage in microns/sec
        self.speed = 1000  # None

        #: str: Resolution of the stage. High or Low.
        self.resolution = "high"  # None

        #: bool: Wait until the stage is done moving before returning
        self.wait_until_done = True

        #: float: Time to wait between checking if the stage is done moving
        self.wait_time = 0.002

        #: int: Number of times to check if the stage is done moving
        self.n_waits = max(int(timeout / self.wait_time), 1)

        # Thread blocking here to prevent calls to get_current_position()
        # while move_to_specified_position is waiting for a response. Serial
        # commands must complete or the MP-285A completely locks up and has
        # to be power cycled.

        #: threading.Event: Event to prevent writing to the serial port
        self.safe_to_write = threading.Event()
        self.safe_to_write.set()

        #: threading.Lock: Lock to prevent writing to the serial port
        self.write_done_flag = threading.Lock()

        #: bool: Flag to indicate if the stage is moving
        self.is_moving = False

        #: time.time: Time of the last write to the serial port
        self.last_write_time = time.time()

        #: int: Number of commands to buffer
        self.commands_num = 10

        #: int: Index of the top command in the buffer
        self.top_command_idx = 0

        #: int: Index of the last command in the buffer
        self.last_command_idx = 0

        #: bool: Flag to indicate if the stage is interrupted
        self.is_interrupted = False

        #: bool: Flat to indicate of the stage is moving.
        self.is_moving = False

        #: int: Number of unreceived bytes.
        self.unreceived_bytes = 0

        #: list: Buffer to store the number of bytes to read for each command
        self.commands_buffer = [1] * self.commands_num

    def send_command(self, command: bytes, response_num=1) -> int:
        """Send a command to the MP-285 stage.

        Parameters
        ----------
        command : bytes
            Command to send to the MP-285 stage.
        response_num : int
            Number of bytes to read for the response.

        Returns
        -------
        idx : int
            Index of the command in the buffer.
        """
        self.safe_to_write.wait()
        self.safe_to_write.clear()
        self.write_done_flag.acquire()
        if self.top_command_idx == self.last_command_idx:
            waiting_bytes = min(self.serial.in_waiting, self.unreceived_bytes)
            if waiting_bytes > 0:
                self.serial.read(waiting_bytes)
                self.unreceived_bytes -= waiting_bytes
                # self.n_waits -= 5
        self.serial.write(command)
        logger.debug(f"MP285 send command {command}")
        self.commands_buffer[self.last_command_idx] = response_num
        idx = self.last_command_idx
        self.last_command_idx = (self.last_command_idx + 1) % self.commands_num
        self.write_done_flag.release()
        return idx

    def read_response(self, idx: int) -> Union[bytes, str, None]:
        """Read the response from the MP-285 stage.

        Parameters
        ----------
        idx : int
            Index of the command in the buffer.

        Returns
        -------
        response : bytes, str, None
            Response from the MP-285 stage.
        """
        if idx != self.top_command_idx:
            return None

        l = self.commands_buffer[self.top_command_idx]  # noqa
        r = ""
        for _ in range(self.n_waits):
            if self.unreceived_bytes > 0 and self.serial.in_waiting > l:
                unreceived = min(self.unreceived_bytes, self.serial.in_waiting - l)
                r = self.serial.read(l + unreceived)
                self.unreceived_bytes -= unreceived
                logger.debug(f"MP285 read response {r}")
                start, end = 0, l + unreceived
                while unreceived > 0:
                    if r[start] == 13:  # b'\r'
                        start += 1
                        unreceived -= 1
                while unreceived > 0 and end - 2 >= start:
                    if r[end - 1] == 13 and r[end - 2] == 13:
                        end -= 1
                        unreceived -= 1
                r = r[start:end]
                logger.debug(f"MP285 valid response: {r}")
                self.n_waits -= 5
                break
            elif self.serial.in_waiting == l:
                r = self.serial.read(l)
                logger.debug(f"MP285 read response {r}")
                break
            time.sleep(self.wait_time)

        if r == "":
            logger.error(
                "Haven't received any responses from MP285! "
                "Please check the stage device!"
            )
            self.unreceived_bytes += self.commands_buffer[self.top_command_idx]
            logger.error(f"MP285 unreceived bytes: {self.unreceived_bytes}")
            # let the waiting time a little bit longer
            self.n_waits += 5
        self.top_command_idx = (self.top_command_idx + 1) % self.commands_num
        self.safe_to_write.set()
        return r
        # raise TimeoutError("Haven't received any responses
        # from MP285! Please check the stage device!")

    def connect_to_serial(self) -> None:
        """Connect to the serial port of the MP-285 stage.

        Raises
        ------
        serial.SerialException
            If the serial connection fails.
        """
        try:
            self.serial.open()
        except serial.SerialException as e:
            print("MP285 serial connection failed.")
            logger.error(f"{str(self)}, Could not open port {self.serial.port}")
            raise e

    def disconnect_from_serial(self) -> None:
        """Disconnect from the serial port of the MP-285 stage."""
        self.serial.close()

    @staticmethod
    def convert_microsteps_to_microns(microsteps: float) -> float:
        """Converts microsteps to microns

        Parameters
        ----------
        microsteps : float
            Distance in microsteps

        Returns
        ---------
        microns : float
            Distance in microns
        """

        microns = np.multiply(microsteps, 0.04)
        return microns

    @staticmethod
    def convert_microns_to_microsteps(microns: float) -> float:
        """Converts microsteps to microns.

        Parameters
        ----------
        microns : float
            Distance in microns.

        Returns
        -------
        microsteps : float
            Distance in microsteps
        """

        microsteps = np.divide(microns, 0.04)
        return microsteps

    def get_current_position(
        self,
    ) -> Tuple[Optional[float], Optional[float], Optional[float]]:
        """Get the current stage position.

        Gets the stage position. The data returned consists of 13 bytes:
        12 bytes containing X, Y, & Z position values in microsteps (4 bytes each),
        followed with the task-complete indicator (1 byte).

        Returns
        -------
        position : Tuple[Optional[float], Optional[float], Optional[float]]
            The x, y, and z positions in microns if available.

            - x_pos is the X position in microns. None if the position is not available.
            - y_pos is the Y position in microns. None if the position is not available.
            - z_pos is the Z position in microns. None if the position is not available.

        """
        # print("calling get_current_position")
        # self.flush_buffers()
        command = bytes.fromhex("63") + bytes.fromhex("0d")
        idx = self.send_command(command, 13)

        while True:
            position_information = self.read_response(idx)
            if position_information is not None:
                break
            time.sleep(self.wait_time)

        if len(position_information) < 13:
            return None, None, None
        xs = int.from_bytes(position_information[0:4], byteorder="little", signed=True)
        ys = int.from_bytes(position_information[4:8], byteorder="little", signed=True)
        zs = int.from_bytes(position_information[8:12], byteorder="little", signed=True)
        x_pos = self.convert_microsteps_to_microns(xs)
        y_pos = self.convert_microsteps_to_microns(ys)
        z_pos = self.convert_microsteps_to_microns(zs)

        return x_pos, y_pos, z_pos

    def move_to_specified_position(
        self, x_pos: float, y_pos: float, z_pos: float
    ) -> bool:
        """Move to Specified Position (‘m’) Command

        This command instructs the controller to move all three axes to the position
        specified.

        The command sequence consists of 14 bytes: Command byte followed by three
        sets of four bytes containing position information in microsteps for X, Y,
        and Z, and ending with the terminator.
        Return data consists of 1 byte (task-complete indicator), which occurs after
        the  move is complete.

        Parameters
        ----------
        x_pos : float
            X position in microns
        y_pos : float
            Y position in microns
        z_pos : float
            Z position in microns

        Returns
        -------
        move_complete : bool
            True if move was successful, False if not.
        """
        # Convert microns to microsteps and create command.
        x_target = int(self.convert_microns_to_microsteps(x_pos))
        y_target = int(self.convert_microns_to_microsteps(y_pos))
        z_target = int(self.convert_microns_to_microsteps(z_pos))
        x_steps = x_target.to_bytes(length=4, byteorder="little", signed=True)
        y_steps = y_target.to_bytes(length=4, byteorder="little", signed=True)
        z_steps = z_target.to_bytes(length=4, byteorder="little", signed=True)

        # Move stage
        move_cmd = (
            bytes.fromhex("6d") + x_steps + y_steps + z_steps + bytes.fromhex("0d")
        )

        idx = self.send_command(move_cmd)
        self.is_moving = True
        while True:
            r = self.read_response(idx)
            if r is not None:
                break
            time.sleep(self.wait_time)
        self.is_moving = False

        return r == bytes.fromhex("0d")

    def set_resolution_and_velocity(self, speed: int, resolution: str) -> bool:
        """Sets the MP-285 stage speed and resolution.

        This command instructs the controller to move all three axes to the position
        specified. The command sequence consists of 4 bytes: Command byte, followed
        by 2 bytes containing resolution and velocity information, and ending with
        the  terminator. Return data consists of 1 byte (task-complete indicator).

        Parameters
        ----------
        speed : int
            Low Resolution = 0-3000 microns/sec
            High Resolution = 0-1310 microns/sec
        resolution : str
            Low - 0.2 microns/microstep = 10 microsteps/step
            High - 0.04 microns/microstep = 50 microsteps/step

        Returns
        -------
        command_complete : bool
            True if command was successful, False if not.
        """
        # print("calling set_resolution_and_velocity")
        # print(f"resolution: {resolution}")
        if resolution == "high":
            resolution_bit = 1
            if speed > 1310:
                speed = 1310
                logger.error("Speed for the high-resolution mode is too fast.")
                raise UserWarning(
                    "High resolution mode of Sutter MP285 speed too "
                    "high. Setting to 1310 microns/sec."
                )
        elif resolution == "low":
            resolution_bit = 0
            if speed > 3000:
                speed = 3000
                logger.error("Speed for the low-resolution mode is too fast.")
                raise UserWarning(
                    "Low resolution mode of Sutter MP285 speed too "
                    "high. Setting to 3000 microns/sec."
                )
        else:
            logger.error("MP-285 resolution must be 'high' or 'low'")
            raise UserWarning("MP-285 resolution must be 'high' or 'low'")

        speed_and_res = int(resolution_bit * 32768 + speed)
        command = (
            bytes.fromhex("56")
            + speed_and_res.to_bytes(length=2, byteorder="little", signed=False)
            + bytes.fromhex("0d")
        )

        # Write Command and get response
        idx = self.send_command(command, 1)
        while True:
            response = self.read_response(idx)
            if response is not None:
                break
            time.sleep(self.wait_time)
        if response == bytes.fromhex("0d"):
            self.speed = speed
            self.resolution = resolution
            command_complete = True
        else:
            command_complete = False

        return command_complete

    def interrupt_move(self) -> Union[bool, None]:
        """Interrupt stage movement.

        This command interrupts and stops a move in progress that originally
        initiated by the Move (‘m’) command. The command sequence consists of 1 byte:
        Command byte (no terminator).

        Return data consists of 1 byte if movement is not in progress, or 2 bytes
        ('=' (move-in- progress indicator)and task-complete indicator.

        Returns
        -------
        stage_stopped : bool
            True if move was successful, False if not.
        """
        if not self.is_moving or self.is_interrupted:
            return True

        # send commands: interrupt and get position
        if self.serial.in_waiting == 0:
            self.is_interrupted = True
            self.write_done_flag.acquire()
            self.serial.write(bytes.fromhex("03630d"))
            logger.debug("MP285 write command 03630d")
            idx = (self.top_command_idx - 1) % self.commands_num
            self.commands_buffer[idx] = 14
            self.top_command_idx = idx
            self.write_done_flag.release()

            while True:
                position_information = self.read_response(idx)
                if position_information is not None:
                    break
                time.sleep(self.wait_time)

            if len(position_information) < 14:
                logger.error(
                    "MP285 didn't get full position information after interruption"
                )
            self.is_interrupted = False

        return True

    def set_absolute_mode(self) -> bool:
        """Set MP285 to Absolute Position Mode.

        This command sets the nature of the positional values specified with the Move
        (‘m’) command as absolute positions as measured from the point of origin
        (Position 0). The command sequence consists of 2 bytes: Command byte, followed
        by the terminator. Return data consists of 1 byte (task-complete indicator).

        Returns
        -------
        command_complete : bool
            True if command was successful, False if not.

        """
        # print("calling set_absolute_mode")
        # self.flush_buffers()
        abs_cmd = bytes.fromhex("61") + bytes.fromhex("0d")
        idx = self.send_command(abs_cmd)
        while True:
            r = self.read_response(idx)
            if r is not None:
                break
            time.sleep(self.wait_time)
        return False

    # def set_relative_mode(self):
    #     """Set MP285 to Relative Position Mode.
    #
    #     This command sets the nature of the positional values specified with the Move
    #     (‘m’) command as relative positions as measured from the current position
    #     (absolute position returned by the Get Current Position (‘c’) command).
    #     The command sequence consists of 2 bytes: Command byte, followed by the
    #     terminator. Return data consists of 1 byte (task-complete indicator).
    #
    #     Returns
    #     -------
    #     command_complete : bool
    #         True if command was successful, False if not.
    #     """
    #     # print("calling set_relative_mode")
    #     self.flush_buffers()
    #     self.safe_write(bytes.fromhex("62") + bytes.fromhex("0d"))
    #     response = self.serial.read(1)
    #     if response == bytes.fromhex("0d"):
    #         command_complete = True
    #     else:
    #         command_complete = False
    #     self.safe_to_write.set()
    #     return command_complete

    def refresh_display(self) -> bool:
        """Refresh the display on the MP-285 controller.

        This command refreshes the VFD (Vacuum Fluorescent Display) of the controller.
        The command sequence consists of 2 bytes: Command byte and terminator.
        Return data consists of 1 byte: task-complete indicator.

        Returns
        -------
        command_complete : bool
            True if command was successful, False if not.
        """
        # print("calling refresh_display")
        # self.flush_buffers()
        idx = self.send_command(bytes.fromhex("6E") + bytes.fromhex("0d"))
        while True:
            response = self.read_response(idx)
            if response is not None:
                break
            time.sleep(self.wait_time)

        return response == bytes.fromhex("0d")

    def reset_controller(self) -> bool:
        """Reset the MP-285 controller.

        This command resets the controller. The command sequence consists of 2 bytes:
        Command byte and terminator. Return data consists of 1 byte: task-complete
        indicator.

        Returns
        -------
        command_complete : bool
            True if command was successful, False if not.
        """
        idx = self.send_command(bytes.fromhex("72") + bytes.fromhex("0d"))

        response = self.read_response(idx)
        if response == bytes.fromhex("0d"):
            command_complete = True
        else:
            command_complete = False
        return command_complete

    def get_controller_status(self) -> bool:
        """Get the status of the MP-285 controller.

        This command gets status information from the controller and returns it in
        fixed-sized block of data. The command sequence consists of 2 bytes: Command
        byte and terminator. Return data consists of 33 bytes: 32 bytes of
        information and task-complete indicator.

        Returns
        -------
        command_complete = bool
            True if command was successful, False if not.
        """
        # print("calling get_controller_status")
        # self.flush_buffers()
        idx = self.send_command(
            bytes.fromhex("73") + bytes.fromhex("0d"), response_num=33
        )
        response = self.read_response(idx)
        if len(response) == 33 and response[-1] == bytes.fromhex("0d"):
            command_complete = True
        else:
            command_complete = False

        # not implemented yet. See page 74 of documentation.
        return command_complete

    def close(self) -> None:
        """Close the serial connection to the stage"""
        self.serial.close()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
import datetime
from typing import Optional, Dict, Any

# Third Party Imports
import numpy as np
import tkinter as tk

import navigate

# Local Imports
from navigate.controller.sub_controllers.gui import GUIController
from navigate.controller.sub_controllers.channels_settings import (
    ChannelSettingController,
)
from navigate.controller.sub_controllers.tiling import TilingWizardController
from navigate.view.main_window_content.channels_tab import ChannelsTab
from navigate.view.popups.tiling_wizard_popup2 import TilingWizardPopup

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ChannelsTabController(GUIController):
    """Controller for the channels tab in the main window."""

    def __init__(
        self,
        view: ChannelsTab,
        parent_controller: Optional["navigate.controller.controller.Controller"],
    ) -> None:
        """Initialize the ChannelsTabController.

        Parameters
        ----------
        view : ChannelsTab
            The ChannelsTab Window.
        parent_controller : navigate.controller.controller.Controller
            The parent controller of this controller.
        """
        super().__init__(view, parent_controller)

        #: bool: Whether the user has selected to save the data.
        self.is_save = False

        #: str: The current acquisition mode.
        self.mode = "stop"

        #: bool: Whether the controller is in the initialization phase.
        self.in_initialization = True

        # sub-controllers
        self.channel_setting_controller = ChannelSettingController(
            self.view.channel_widgets_frame,
            self,
            self.parent_controller.configuration_controller,
        )

        self.view.stack_timepoint_frame.exp_time_spinbox.set_precision(0)

        #: dict: The widgets in the stack acquisition settings frame.
        self.stack_acq_widgets = self.view.stack_acq_frame.get_widgets()

        #: dict: The values in the stack acquisition settings frame.
        self.stack_acq_vals = self.view.stack_acq_frame.get_variables()

        #: dict: The buttons in the stack acquisition settings frame.
        self.stack_acq_buttons = self.view.stack_acq_frame.get_buttons()

        # stack acquisition event binds
        self.stack_acq_vals["step_size"].trace_add("write", self.update_z_steps)
        self.stack_acq_vals["start_position"].trace_add("write", self.update_z_steps)
        self.stack_acq_vals["end_position"].trace_add("write", self.update_z_steps)
        self.stack_acq_vals["start_focus"].trace_add("write", self.update_z_steps)
        self.stack_acq_vals["z_device"].trace_add("write", self.update_additional_stacking_axes)
        self.stack_acq_vals["f_device"].trace_add("write", self.update_additional_stacking_axes)
        self.stack_acq_buttons["set_start"].configure(
            command=self.update_start_position
        )
        self.stack_acq_buttons["set_end"].configure(command=self.update_end_position)

        # stack acquisition_variables
        #: float: The z origin of the stack.
        self.z_origin = 0

        #: float: The focus origin of the stack.
        self.focus_origin = 0

        #: float: The filter wheel delay.
        self.filter_wheel_delay = None

        #: dict: The microscope state dictionary.
        self.microscope_state_dict = {}

        # laser/stack cycling event binds
        self.stack_acq_vals["cycling"].trace_add("write", self.update_cycling_setting)

        # time point setting variables
        temp = self.view.stack_timepoint_frame

        #: dict: Dictionary of time point settings.
        self.timepoint_vals = {
            "is_save": temp.save_data,
            "timepoints": temp.exp_time_spinval,
            "stack_acq_time": temp.stack_acq_spinval,
            "stack_pause": temp.stack_pause_spinval,
            "experiment_duration": temp.total_time_spinval,
            "timepoint_interval": temp.timepoint_interval_spinval,
        }

        # time point event binds
        self.timepoint_vals["is_save"].trace_add("write", self.update_save_setting)
        self.timepoint_vals["timepoints"].trace_add(
            "write", lambda *args: self.update_timepoint_setting()
        )
        self.timepoint_vals["stack_pause"].trace_add(
            "write", lambda *args: self.update_timepoint_setting()
        )

        # Multi Position Acquisition
        #: bool: Whether the user has selected to use multi-position.
        self.is_multiposition = False

        #: bool: cache multi-position flag
        self.is_multiposition_cache = False

        #: bool: Whether the user has selected to use multi-position.
        self.is_multiposition_val = self.view.multipoint_frame.on_off
        self.is_multiposition_val.trace_add("write", self.toggle_multiposition)

        self.view.multipoint_frame.buttons["tiling"].configure(
            command=self.launch_tiling_wizard
        )

        # Waveform Parameters
        self.view.quick_launch.buttons["waveform_parameters"].configure(
            command=self.launch_waveform_parameters
        )

        # Autofocus Settings
        self.view.quick_launch.buttons["autofocus_button"].configure(
            command=self.launch_autofocus_settings
        )

        self.initialize()

    def initialize(self) -> None:
        """Initializes widgets and gets other necessary configuration.

        The `initialize` method in the `ChannelsTabController` class is responsible
        for setting up the initial configuration and state of the channels tab in the
        main window.
        """
        config = self.parent_controller.configuration_controller
        self.stack_acq_widgets["cycling"].widget["values"] = ["Per Z", "Per Stack"]

        # Set the default stage for acquiring a z-stack.
        z_stages = config.get_stages_by_axis("z")
        self.stack_acq_widgets["z_device"].widget["values"] = z_stages
        if len(z_stages) >= 1:
            self.stack_acq_widgets["z_device"].widget.current(0)

        f_stages = config.get_stages_by_axis("f")
        self.stack_acq_widgets["f_device"].widget["values"] = f_stages
        if len(f_stages) >= 1:
            self.stack_acq_widgets["f_device"].widget.current(0)

        axes = config.stage_axes
        axes = [axis for axis in axes if axis not in ("x", "y", "theta")]
        devices_dict = {}
        for axis in axes:
            temp = config.get_stages_by_axis(axis)
            for device in temp:
                device_name, axis = device.split(" - ")
                devices_dict[axis] = device_name
        self.view.stack_acq_frame.create_additional_stack_widgets(axes, devices_dict)


        self.filter_wheel_delay = [
            config.filter_wheel_setting_dict[i]["filter_wheel_delay"]
            for i in range(config.number_of_filter_wheels)
        ]
        self.channel_setting_controller.initialize()
        self.set_spinbox_range_limits(self.parent_controller.configuration["gui"])
        self.show_verbose_info("channels tab has been initialized")

    def populate_experiment_values(self) -> None:
        """Distribute initial MicroscopeState values to this and sub-controllers and
        associated views.

        The populate_experiment_values method in the ChannelsTabController class is
        responsible for distributing initial MicroscopeState values to this and
        sub-controllers and associated views. It sets the initial values for various
        settings, validates the configuration for multi-position settings,
        and updates the GUI accordingly.

        """
        self.in_initialization = True
        self.microscope_state_dict = self.parent_controller.configuration["experiment"][
            "MicroscopeState"
        ]
        if self.microscope_state_dict["step_size"] < 0:
            self.microscope_state_dict["step_size"] = -self.microscope_state_dict[
                "step_size"
            ]
        self.set_info(self.stack_acq_vals, self.microscope_state_dict)
        self.set_info(self.timepoint_vals, self.microscope_state_dict)

        # set advanced stack acquistion settings
        for axis in ["z", "f"]:
            devices = self.stack_acq_widgets[f"{axis}_device"].widget["values"]
            idx = 0
            for i, device in enumerate(devices):
                if device.endswith(self.microscope_state_dict.get(f"primary_{axis}_axis", axis)):
                    idx = i
                    break
            self.stack_acq_widgets[f"{axis}_device"].widget.current(idx)

        secondary_stack_settings= self.microscope_state_dict.get("secondary_stack_settings", {})
        variable_dict = self.view.stack_acq_frame.additional_stack_setting_variables
        for axis in secondary_stack_settings.keys():
            index_axis = f"stack_{axis}"
            if index_axis in variable_dict:
                variable_dict[index_axis].set(True)
                self.view.stack_acq_frame.update_setting_widgets(axis)()
                variable_dict[f"{axis}_offset"].set(secondary_stack_settings[axis])
        
        self.update_additional_stacking_axes()

        # check configuration for multi-position settings
        self.is_multiposition_val.set(self.microscope_state_dict["is_multiposition"])
        self.is_multiposition_cache = self.is_multiposition
        self.toggle_multiposition()

        # validate
        self.view.stack_timepoint_frame.stack_pause_spinbox.trigger_focusout_validation()
        self.view.stack_timepoint_frame.exp_time_spinbox.trigger_focusout_validation()

        if self.microscope_state_dict["stack_cycling_mode"] not in [
            "per_z",
            "per_stack",
        ]:
            self.microscope_state_dict["stack_cycling_mode"] = "per_stack"
        self.stack_acq_vals["cycling"].set(
            "Per Z"
            if self.microscope_state_dict["stack_cycling_mode"] == "per_z"
            else "Per Stack"
        )
        self.channel_setting_controller.populate_experiment_values(
            self.microscope_state_dict["channels"]
        )

        # after initialization
        self.in_initialization = False
        self.channel_setting_controller.in_initialization = False
        # update z and f position
        self.z_origin = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]["z"]
        self.focus_origin = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]["f"]
        self.update_z_steps()

        self.show_verbose_info("Channels tab has been set new values")

    def set_spinbox_range_limits(self, settings: Dict[str, Any]) -> None:
        """Sets range limits for various spinbox widgets based on the provided settings.

        This method configures the minimum, maximum, and increment values for the spinbox
        widgets in the stack acquisition settings frame and the time point settings
        frame based on the provided configuration settings.

        Parameters
        ----------
        settings : Dict[str, Any]
            dictionary of settings from configuration file
        """

        # Z-Stack Step Size
        self.stack_acq_widgets["step_size"].widget.configure(
            from_=settings["stack_acquisition"]["step_size"].get("min", 0.01),
            to=settings["stack_acquisition"]["step_size"].get("max", 1000),
            increment=settings["stack_acquisition"]["step_size"].get("step", 0.01),
        )

        # Z-Stack Z Start Position
        self.stack_acq_widgets["start_position"].widget.configure(
            from_=settings["stack_acquisition"]["z_start_pos"].get("min", -1000),
            to=settings["stack_acquisition"]["z_start_pos"].get("max", 1000),
            increment=settings["stack_acquisition"]["z_start_pos"].get("step", 1),
        )

        # Z-Stack Z End Position
        self.stack_acq_widgets["end_position"].widget.configure(
            from_=settings["stack_acquisition"]["z_end_pos"].get("min", -1000),
            to=settings["stack_acquisition"]["z_end_pos"].get("max", 1000),
            increment=settings["stack_acquisition"]["z_end_pos"].get("step", 1),
        )

        # Z-Stack F Start Position
        self.stack_acq_widgets["start_focus"].widget.configure(
            from_=settings["stack_acquisition"]["f_start_pos"].get("min", -1000),
            to=settings["stack_acquisition"]["f_start_pos"].get("max", 1000),
            increment=settings["stack_acquisition"]["f_start_pos"].get("step", 1),
        )

        # Z-Stack F End Position
        self.stack_acq_widgets["end_focus"].widget.configure(
            from_=settings["stack_acquisition"]["f_end_pos"].get("min", -1000),
            to=settings["stack_acquisition"]["f_end_pos"].get("max", 1000),
            increment=settings["stack_acquisition"]["f_end_pos"].get("step", 1),
        )

        # Stack Pause Duration
        self.view.stack_timepoint_frame.stack_pause_spinbox.configure(
            from_=settings["time"]["stack_pause"].get("min", 0),
            to=settings["time"]["stack_pause"].get("max", 100),
            increment=settings["time"]["stack_pause"].get("step", 1),
        )

        # Timepoints
        self.view.stack_timepoint_frame.exp_time_spinbox.configure(
            from_=settings["time"]["timepoints"].get("min", 1),
            to=settings["time"]["timepoints"].get("max", 5000),
            increment=settings["time"]["timepoints"].get("step", 1),
        )

        # Channel settings
        self.channel_setting_controller.set_spinbox_range_limits(settings)

    def set_mode(self, mode: str) -> None:
        """Change acquisition mode.

        The set_mode method changes the acquisition mode and updates the state of
        various widgets accordingly

        Parameters
        ----------
        mode : str
            acquisition mode
        """

        # image_mode: imaging mode, e.g., "live", "single", "z-stack", "customized"
        # stack acquisition settings are disabled in "live" and "single" mode.
        image_mode = self.microscope_state_dict["image_mode"]
        self.mode = mode
        self.channel_setting_controller.set_mode(mode)

        # state_readonly = "readonly" if mode == "stop" else "disabled"
        if mode != "stop":
            state = "disabled"
        elif image_mode == "live" or image_mode == "single":
            state = "disabled"
        else:
            state = "normal"
        for widget_name in [
            "start_position",
            "start_focus",
            "end_position",
            "end_focus",
            "step_size",
        ]:
            self.stack_acq_widgets[widget_name].widget["state"] = state
        for widget_name in ["cycling", "z_device", "f_device"]:
            self.stack_acq_widgets[widget_name].widget["state"] = (
                "readonly" if state == "normal" else "disabled"
            )
        for widget_name in self.view.stack_acq_frame.additional_stack_setting_variables:
            if widget_name.startswith("stack_"):
                self.stack_acq_widgets[widget_name]["state"] = state
            elif widget_name.endswith("_offset"):
                self.stack_acq_widgets[widget_name]["state"] = state
        if state == "normal":
            self.update_additional_stacking_axes()

        self.view.stack_timepoint_frame.save_check["state"] = (
            "normal" if image_mode == "single" and mode == "stop" else state
        )
        self.view.stack_timepoint_frame.stack_pause_spinbox["state"] = state
        self.view.stack_timepoint_frame.exp_time_spinbox["state"] = state

        # multi-position flag
        if mode == "stop":
            self.is_multiposition_val.set(self.is_multiposition_cache)
        else:
            self.is_multiposition_cache = self.is_multiposition
            if mode == "customized":
                self.is_multiposition_val.set(False)

        if image_mode == "customized" or mode != "stop":
            self.disable_multiposition_btn()
        else:
            self.enable_multiposition_btn()

        self.show_verbose_info("acquisition mode has been changed to", mode)

    def update_z_steps(self, *_: tuple[str]) -> None:
        """Recalculates the number of slices that will be acquired in a z-stack.

        Requires GUI to have start position, end position, or step size changed.
        Sets the number of slices in the model and the GUI. Sends the current values
        to central/parent controller

        Parameters
        ----------
        _ : tuple[str]
            Values is a tuple of strings. e.g., ('PY_VAR0', '', 'write')
        """

        # won't do any calculation when initialization
        if self.in_initialization:
            return

        # Calculate the number of slices and set GUI
        try:
            # validate the spin box's value
            start_position = float(self.stack_acq_vals["start_position"].get())
            end_position = float(self.stack_acq_vals["end_position"].get())
            step_size = float(self.stack_acq_vals["step_size"].get())
            if step_size < 0.001:
                self.stack_acq_vals["number_z_steps"].set(0)
                self.microscope_state_dict["abs_z_start"] = 0
                self.microscope_state_dict["abs_z_end"] = 0
                # self.stack_acq_vals["abs_z_start"].set(0)
                # self.stack_acq_vals["abs_z_end"].set(0)
                return
        except tk.TclError:
            self.stack_acq_vals["number_z_steps"].set(0)
            self.microscope_state_dict["abs_z_start"] = 0
            self.microscope_state_dict["abs_z_end"] = 0
            # self.stack_acq_vals["abs_z_start"].set(0)
            # self.stack_acq_vals["abs_z_end"].set(0)
            return
        except (KeyError, AttributeError):
            logger.error("Error caught: updating z_steps")
            return

        number_z_steps = int(
            np.ceil(np.abs((end_position - start_position) / step_size))
        )
        self.stack_acq_vals["number_z_steps"].set(number_z_steps)

        # Shift the start/stop positions by the relative position
        flip_flags = self.parent_controller.configuration_controller.stage_flip_flags
        if flip_flags["z"]:
            self.microscope_state_dict["abs_z_start"] = self.z_origin + end_position
            self.microscope_state_dict["abs_z_end"] = self.z_origin + start_position
            # self.stack_acq_vals["abs_z_start"].set(self.z_origin + end_position)
            # self.stack_acq_vals["abs_z_end"].set(self.z_origin + start_position)
        else:
            self.microscope_state_dict["abs_z_start"] = self.z_origin + start_position
            self.microscope_state_dict["abs_z_end"] = self.z_origin + end_position
            # self.stack_acq_vals["abs_z_start"].set(self.z_origin + start_position)
            # self.stack_acq_vals["abs_z_end"].set(self.z_origin + end_position)

        # update experiment MicroscopeState dict
        self.microscope_state_dict["start_position"] = start_position
        self.microscope_state_dict["end_position"] = end_position
        self.microscope_state_dict["step_size"] = step_size * (
            -1 if flip_flags["z"] else 1
        )
        self.microscope_state_dict["number_z_steps"] = number_z_steps
        self.stack_acq_vals["bottom"].set(
            self.microscope_state_dict["abs_z_start"]
        )
        self.stack_acq_vals["top"].set(
            self.microscope_state_dict["abs_z_end"]
        )

        try:
            self.microscope_state_dict["start_focus"] = self.stack_acq_vals[
                "start_focus"
            ].get()
        except tk.TclError:
            self.microscope_state_dict["start_focus"] = 0
        try:
            self.microscope_state_dict["end_focus"] = self.stack_acq_vals[
                "end_focus"
            ].get()
        except tk.TclError:
            self.microscope_state_dict["end_focus"] = 0
        self.microscope_state_dict["stack_z_origin"] = self.z_origin
        self.microscope_state_dict["stack_focus_origin"] = self.focus_origin

        self.update_timepoint_setting()
        self.show_verbose_info(
            "stack acquisition settings on channels tab have been changed and "
            "recalculated"
        )

    def update_start_position(self, *_: tuple[str]) -> None:
        """Get new z starting position from current stage parameters.

        Parameters
        ----------
        _ : tuple[str]
            Values is a tuple of strings. e.g., ('PY_VAR0', '', 'write')
        """

        # We have a new origin
        self.z_origin = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]["z"]
        self.focus_origin = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]["f"]

        flip_flags = self.parent_controller.configuration_controller.stage_flip_flags
        if flip_flags["z"]:
            self.stack_acq_vals["end_position"].set(0)
            self.stack_acq_vals["end_focus"].set(0)
        else:
            self.stack_acq_vals["start_position"].set(0)
            self.stack_acq_vals["start_focus"].set(0)

        # Propagate parameter changes to the GUI
        self.update_z_steps()

    def update_end_position(self, *_: tuple[str]) -> None:
        """Get new z ending position from current stage parameters

        Parameters
        ----------
        _ : tuple[str]
            Values is a tuple of strings. e.g., ('PY_VAR0', '', 'write')

        """
        # Grab current values
        z_end = self.parent_controller.configuration["experiment"]["StageParameters"][
            "z"
        ]
        focus_end = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]["f"]

        z_start = self.z_origin
        focus_start = self.focus_origin

        if z_end < z_start:
            # Sort so we are always going low to high
            z_start, z_end = z_end, z_start
            focus_start, focus_end = focus_end, focus_start

        # set origin to be in the middle of start and end
        self.z_origin = (z_start + z_end) / 2
        self.focus_origin = (focus_start + focus_end) / 2

        # Propagate parameter changes to the GUI
        flip_flags = self.parent_controller.configuration_controller.stage_flip_flags
        start_pos = z_start - self.z_origin
        end_pos = z_end - self.z_origin
        start_focus = focus_start - self.focus_origin
        end_focus = focus_end - self.focus_origin
        if flip_flags["z"]:
            start_pos, end_pos = end_pos, start_pos
            start_focus, end_focus = end_focus, start_focus
        self.stack_acq_vals["start_position"].set(start_pos)
        self.stack_acq_vals["start_focus"].set(start_focus)
        self.stack_acq_vals["end_position"].set(end_pos)
        self.stack_acq_vals["end_focus"].set(end_focus)
        self.update_z_steps()

    def update_cycling_setting(self, *_: tuple[str]) -> None:
        """Update the cycling settings in the model and the GUI.

        You can collect different channels in different formats.
        In the perZ format: Slice 0/Ch0, Slice0/Ch1, Slice1/Ch0, Slice1/Ch1, etc.
        in the perStack format: Slice 0/Ch0, Slice1/Ch0... SliceN/Ch0.  Then it repeats
        with Ch1

        Parameters
        ----------
        _ : tuple[str]
            Values is a tuple of strings. e.g., ('PY_VAR0', '', 'write')
        """

        # won't do any calculation when initializing
        if self.in_initialization:
            return
        # update experiment MicroscopeState dict
        self.microscope_state_dict["stack_cycling_mode"] = (
            "per_stack"
            if self.stack_acq_vals["cycling"].get() == "Per Stack"
            else "per_z"
        )

        # recalculate time point settings
        self.update_timepoint_setting()

        self.show_verbose_info("Cycling setting on channels tab has been changed")

    def update_save_setting(self, *_: tuple[str]) -> None:
        """Tell the parent controller 'save_data' is selected.

        Does not do any calculation when initializing the software.

        Parameters
        ----------
        _ : tuple[str]
            Values is a tuple of strings. e.g., ('PY_VAR0', '', 'write')
        """

        if self.in_initialization:
            return
        self.is_save = self.timepoint_vals["is_save"].get()
        self.microscope_state_dict["is_save"] = self.is_save
        self.parent_controller.execute("set_save", self.is_save)
        self.show_verbose_info("Save data option has been changed to", self.is_save)

    def update_timepoint_setting(self) -> None:
        """Automatically calculates the stack acquisition time based on the number of
        time points, channels, and exposure time.

        TODO: Add necessary computation for 'Stack Acq.Time', 'Time point Interval',
        'Experiment Duration'?

        Does not do any calculation when initializing the software.
        Order of priority for per_stack: timepoints > positions > channels > z-steps
                                        > delay
        Order of priority for perZ: timepoints > positions > z-steps > delays > channels
        """

        if self.in_initialization:
            return
        channel_settings = self.microscope_state_dict["channels"]
        number_of_positions = (
            self.parent_controller.multiposition_tab_controller.get_position_num()
            if self.is_multiposition
            else 1
        )
        channel_exposure_time = []
        # validate the spin box's value
        try:
            number_of_timepoints = int(float(self.timepoint_vals["timepoints"].get()))
            number_of_slices = int(self.stack_acq_vals["number_z_steps"].get())
            for channel_id in channel_settings.keys():
                channel = channel_settings[channel_id]
                if channel["is_selected"]:
                    channel_exposure_time.append(float(channel["camera_exposure_time"]))
            if len(channel_exposure_time) == 0:
                return
        except (tk.TclError, ValueError):
            self.timepoint_vals["experiment_duration"].set("0")
            self.timepoint_vals["stack_acq_time"].set("0")
            return
        except (KeyError, AttributeError):
            logger.error("Error caught: updating time point setting")
            return

        per_stack = self.stack_acq_vals["cycling"].get() == "Per Stack"

        # Initialize variable to keep track of how long the entire experiment will take.
        # Includes time, positions, channels...
        experiment_duration = 0

        # Initialize variable to calculate how long it takes to acquire a single volume
        # for all the channels. Only calculate once at the beginning.
        stack_acquisition_duration = 0

        for position_idx in range(number_of_positions):
            # For multiple positions, need to account for the time necessary to move
            # the stages that distance. In theory, these positions would be
            # populated in that 'pandastable' or some other data structure.

            # Determine the largest distance to travel between positions.  Assume
            # all axes move the same velocity This assumes that we are in a
            # multi-position mode. Not yet implemented.
            # x1, y1, z1, theta1, f1, = position_start.values()
            # x2, y2, z1, theta2, f1 = position_end.values()
            # distance = [x2-x1, y2-y1, z2-z1, theta2-theta1, f2-f1]
            # max_distance_idx = np.argmax(distance)
            # Now if we are going to do this properly, we would need to do this for
            # all the positions so that we can calculate the total experiment
            # time. Probably assemble a matrix of all the positions and then do
            # the calculations.

            stage_delay = 0
            # TODO False value.

            # If we were actually acquiring the data, we would call the function to
            # move the stage here.
            experiment_duration = experiment_duration + stage_delay

            for channel_idx in range(len(channel_exposure_time)):
                if per_stack:
                    # In the per_stack mode, we only need to account for the time
                    # necessary for the filter wheel to change between each
                    # image stack.
                    if channel_idx == 0 and position_idx == 0:
                        stack_acquisition_duration += (
                            channel_exposure_time[channel_idx] / 1000 * number_of_slices
                        )
                else:
                    if position_idx == 0:
                        stack_acquisition_duration += (
                            channel_exposure_time[channel_idx] / 1000 * number_of_slices
                        )

                experiment_duration += (
                    channel_exposure_time[channel_idx] / 1000 * number_of_slices
                )

            try:
                stack_pause = float(self.timepoint_vals["stack_pause"].get())
            except ValueError:
                stack_pause = 0
            experiment_duration = experiment_duration + stack_pause
        experiment_duration *= number_of_timepoints

        # Change the filter wheel here before the start of the acquisition.
        if len(channel_exposure_time) > 1:
            filter_wheel_change_times = number_of_timepoints * (
                1 if per_stack else number_of_slices
            )
            experiment_duration += (
                sum(self.filter_wheel_delay) * filter_wheel_change_times
            )
        else:
            experiment_duration += sum(self.filter_wheel_delay)
        self.timepoint_vals["experiment_duration"].set(
            str(datetime.timedelta(seconds=experiment_duration))
        )
        self.timepoint_vals["stack_acq_time"].set(
            str(datetime.timedelta(seconds=stack_acquisition_duration))
        )

        # update experiment MicroscopeState dict
        self.microscope_state_dict["timepoints"] = number_of_timepoints
        self.microscope_state_dict["stack_pause"] = self.timepoint_vals[
            "stack_pause"
        ].get()
        self.microscope_state_dict["stack_acq_time"] = stack_acquisition_duration
        self.microscope_state_dict["experiment_duration"] = experiment_duration

        self.show_verbose_info(
            "time point settings on channels tab have been changed and recalculated"
        )

    def toggle_multiposition(self, *_: tuple[str]) -> None:
        """Toggle Multi-position Acquisition.

        Recalculates the experiment duration.

        Parameters
        ----------
        _ : tuple[str]
            Values is a tuple of strings. e.g., ('PY_VAR0', '', 'write')
        """
        self.is_multiposition = self.is_multiposition_val.get()
        self.microscope_state_dict["is_multiposition"] = self.is_multiposition
        self.update_timepoint_setting()
        self.show_verbose_info("Multi-position:", self.is_multiposition)

    def disable_multiposition_btn(self) -> None:
        """Disable multi-position button"""
        self.view.multipoint_frame.save_check.config(state="disabled")

    def enable_multiposition_btn(self) -> None:
        """Enable multi-position button"""
        self.view.multipoint_frame.save_check.config(state="normal")

    def launch_waveform_parameters(self) -> None:
        """Launches waveform parameters popup."""
        self.parent_controller.menu_controller.popup_waveform_setting()

    def launch_autofocus_settings(self) -> None:
        """Launches autofocus settings popup."""
        self.parent_controller.menu_controller.popup_autofocus_setting()

    def launch_tiling_wizard(self) -> None:
        """Launches tiling wizard popup.

        Will only launch when button in GUI is pressed, and will not duplicate.
        Pressing button again brings popup to top
        """

        if hasattr(self, "tiling_wizard_controller"):
            self.tiling_wizard_controller.showup()
            return
        stage_axes = self.parent_controller.configuration_controller.stage_axes
        # not tiling on theta axis right now
        tiling_wizard = TilingWizardPopup(self.view, axes=[axis.upper() for axis in stage_axes if axis != "theta"])
        self.tiling_wizard_controller = TilingWizardController(tiling_wizard, self)

    @staticmethod
    def set_info(vals: Dict[str, Any], values: Dict[str, Any]) -> None:
        """Set values to a list of variables.

        Parameters
        ----------
        vals : Dict[str, Any]
            A dictionary of variables to set.
        values : Dict[str, Any]
            A dictionary of values to set to variables.
        """
        for name in values.keys():
            if name in vals:
                vals[name].set(values[name])

    def execute(self, command: str, *args: tuple[str]) -> None:
        """Execute Command in the parent controller.

        Parameters
        ----------
        command : str
            recalculate_timepoint, channel, move_stage_and_update_info,
            get_stage_position
        args : tuple[str]
            A tuple of arguments to pass to the command.
        """
        if command == "recalculate_timepoint":
            self.update_timepoint_setting()
            # update framerate info in camera setting tab
            exposure_time = max(
                map(
                    lambda channel: (
                        float(channel["camera_exposure_time"])
                        if channel["is_selected"]
                        else 0
                    ),
                    self.microscope_state_dict["channels"].values(),
                )
            )
            self.parent_controller.camera_setting_controller.update_exposure_time(
                exposure_time
            )
        elif (command == "channel") or (command == "update_setting"):
            self.view.after(
                1000, lambda: self.parent_controller.execute(command, *args)
            )

        self.show_verbose_info("Received command from child", command, args)

    def update_experiment_values(self) -> None:
        """Update experiment values"""
        self.channel_setting_controller.update_experiment_values()
        self.update_z_steps()

        # update primary/secondary stack acquisition stage settings
        primary_z_stage = self.stack_acq_vals["z_device"].get()
        primary_z_axis = primary_z_stage.split(" - ")[1]
        primary_f_stage = self.stack_acq_vals["f_device"].get()
        primary_f_axis = primary_f_stage.split(" - ")[1]

        self.microscope_state_dict["primary_z_axis"] = primary_z_axis
        self.microscope_state_dict["primary_f_axis"] = primary_f_axis

        secondary_stack_settings= {}
        variable_dict = self.view.stack_acq_frame.additional_stack_setting_variables
        for k in variable_dict.keys():
            if k.startswith("stack_") and variable_dict[k].get():
                axis = k.split("_")[1]
                offset = variable_dict[f"{axis}_offset"].get()
                secondary_stack_settings[axis] = offset

        self.microscope_state_dict["secondary_stack_settings"] = secondary_stack_settings

    def verify_experiment_values(self) -> str:
        """Verify channel tab settings and return warning info

        Returns
        -------
        string: str
            Warning info
        """
        warning = self.channel_setting_controller.verify_experiment_values()
        if warning:
            return warning
        if self.microscope_state_dict["image_mode"] not in ["live", "single"]:
            if (
                self.microscope_state_dict["number_z_steps"]
                != self.stack_acq_vals["number_z_steps"].get()
            ):
                return "There is something wrong with the stack settings!"
            if self.microscope_state_dict["number_z_steps"] < 1:
                return "The number of Z steps should be at least 1!"
            try:
                float(self.microscope_state_dict["stack_pause"])
            except Exception as e:
                logger.exception(e)
                return "Stack pause should be a valid number!"
            if self.microscope_state_dict["timepoints"] < 1:
                return "Timepoints should be at least 1!"
        return ""

    def set_exposure_time(self, channel_exposure_time: tuple[str, float]) -> None:
        """Set exposure time for a specified channel

        Parameters
        ----------
        channel_exposure_time : tuple(str, float)
            (channel_name, exposure_time)
            Channel name, such as "channel_1", "channel_2",...
            Exposure time in milliseconds.
        """
        channel, exposure_time = channel_exposure_time
        idx = int(channel[channel.index("_") + 1 :]) - 1
        self.channel_setting_controller.in_initialization = True
        self.channel_setting_controller.view.exptime_variables[idx].set(exposure_time)
        self.channel_setting_controller.in_initialization = False

    def update_additional_stacking_axes(self, *args, **kwargs):
        if self.stack_acq_vals["z_device"].get():
            z_axis = self.stack_acq_vals["z_device"].get().split(" - ")[1]
        else:
            z_axis = ""
        if self.stack_acq_vals["f_device"].get():
            f_axis = self.stack_acq_vals["f_device"].get().split(" - ")[1]
        else:
            f_axis = ""

        # enable all axis
        for widget_name in self.view.stack_acq_frame.additional_stack_setting_variables:
            if widget_name.startswith("stack_"):
                self.stack_acq_widgets[widget_name].state(["!disabled"])

        # disable primary z and f
        variable_dict = self.view.stack_acq_frame.additional_stack_setting_variables
        if f"stack_{z_axis}" in variable_dict:
            self.stack_acq_widgets[f"stack_{z_axis}"].state(["disabled"])
            if variable_dict[f"stack_{z_axis}"].get():
                variable_dict[f"stack_{z_axis}"].set(False)
                self.view.stack_acq_frame.update_setting_widgets(z_axis)()
        if f"stack_{f_axis}" in variable_dict:
            self.stack_acq_widgets[f"stack_{f_axis}"].state(["disabled"])
            if variable_dict[f"stack_{f_axis}"].get():
                variable_dict[f"stack_{f_axis}"].set(False)
                self.view.stack_acq_frame.update_setting_widgets(f_axis)()

    @property
    def custom_events(self) -> dict[str, callable]:
        """Custom events for the channels tab."""
        return {"exposure_time": self.set_exposure_time}



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Library Imports
import logging
import time
import traceback

# Third Party Imports
import nidaqmx
from nidaqmx.errors import DaqError
from nidaqmx.constants import LineGrouping

# Local Imports
from navigate.model.devices.filter_wheel.base import FilterWheelBase
from navigate.model.devices.device_types import NIDevice
from navigate.tools.decorators import log_initialization

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class NIFilterWheel(FilterWheelBase, NIDevice):
    """DAQFilterWheel - Class for controlling filter wheels with a DAQ."""

    def __init__(self, microscope_name, device_connection, configuration, device_id):
        """Initialize the DAQFilterWheel class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : Any
            The communication instance with the device.
        configuration : Dict[str, Any]
            Global configuration dictionary.
        device_id : int
            The ID of the device. Default is 0.
        """

        super().__init__(microscope_name, device_connection, configuration, device_id)

        #: float: Delay for filter wheel to change positions.
        self.wait_until_done_delay = self.device_config["filter_wheel_delay"]

        self.filter_wheel_task = None

    def __str__(self):
        """String representation of the class."""
        return "DAQFilterWheel"

    def __enter__(self):
        """Enter the ASI Filter Wheel context manager."""
        return self

    def __exit__(self):
        """Exit the ASI Filter Wheel context manager."""
        if self.filter_wheel_task:
            try:
                self.filter_wheel_task.stop()
                self.filter_wheel_task.close()
            except Exception:
                pass

    def set_filter(self, filter_name, wait_until_done=True):
        """Change the filter wheel to the filter designated by the filter
        position argument. Requires a digital port on the DAQ.

        Parameters
        ----------
        filter_name : str
            Name of filter to move to.
        wait_until_done : bool
            Waits duration of time necessary for filter wheel to change positions.
        """
        if self.check_if_filter_in_filter_dictionary(filter_name) is True:
            try:
                # Create the nidaqmx Task, and add the DO channel.
                self.filter_wheel_task = nidaqmx.Task()
                self.filter_wheel_task.do_channels.add_do_chan(
                    lines=self.filter_dictionary[filter_name],
                    line_grouping=LineGrouping.CHAN_FOR_ALL_LINES,
                )

                # Trigger the nidaqmx Task to a 5V state.
                self.filter_wheel_task.write([True], auto_start=True)

                #  Wheel Position Change Delay in seconds
                if wait_until_done:
                    time.sleep(self.wait_until_done_delay)

                # Trigger the nidaqmx Task to a 0V state.
                self.filter_wheel_task.write([False], auto_start=True)

                # Clean up the task
                self.filter_wheel_task.stop()
                self.filter_wheel_task.close()
            except DaqError as e:
                logger.debug(e)

    def close(self) -> None:
        """Close the DAQ Filter Wheel

        Sets the filter wheel to the home position and then closes the port.
        """
        pass

    def __del__(self) -> None:
        """Delete the DAQFilterWheel object."""
        if self.filter_wheel_task:
            try:
                self.filter_wheel_task.stop()
                self.filter_wheel_task.close()
            except Exception:
                logger.exception(f"Error stopping task: {traceback.format_exc()}")



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from time import sleep
from tkinter import filedialog, messagebox
from typing import Optional

# Third Party Imports

# Local Imports
from navigate.view.configurator_application_window import ConfigurationAssistantWindow
from navigate.view.configurator_application_window import (
    MicroscopeTab,
    MicroscopeWindow,
)
from navigate.config.configuration_database import (
    hardwares_dict,
    hardwares_config_name_dict,
)
from navigate.tools.file_functions import load_yaml_file

# Logger Setup
import logging

p = __name__.split(".")[1]
logger = logging.getLogger(p)


class Configurator:
    """Navigate Configurator"""

    def __init__(self, root: tk.Tk, splash_screen):
        """Initiates the configurator application window.

        Parameters
        ----------
        root : tk.Tk
            The main window of the application
        splash_screen : SplashScreen
            The splash screen of the application
        """
        self.root = root

        # Show the splash screen for 1 second and then destroy it.
        sleep(1)
        splash_screen.destroy()
        self.root.deiconify()
        self.view = ConfigurationAssistantWindow(root)
        self.view.microscope_window = MicroscopeWindow(
            self.view.microscope_frame, self.view.root
        )

        self.view.top_window.add_button.config(command=self.add_microscope)
        self.view.top_window.new_button.config(command=self.new_configuration)
        self.view.top_window.load_button.config(command=self.load_configuration)
        self.view.top_window.save_button.config(command=self.save)
        self.view.top_window.cancel_button.config(command=self.on_cancel)
        self.microscope_id = 0
        self.create_config_window(0)

    def on_cancel(self) -> None:
        """Closes the window and exits the program"""
        self.root.destroy()
        exit()

    def add_microscope(self) -> None:
        """Add a new microscope tab"""
        self.microscope_id += 1
        self.create_config_window(self.microscope_id)

    def delete_microscopes(self) -> None:
        """Delete all microscopes"""
        # delete microscopes
        for tab_id in self.view.microscope_window.tabs():
            self.view.microscope_window.forget(tab_id)
        self.view.microscope_window.tab_list = []
        self.microscope_id = 0

    def new_configuration(self) -> None:
        """Create new configurations"""
        self.delete_microscopes()
        self.create_config_window(self.microscope_id)

    def save(self) -> None:
        """Save configuration file"""

        def set_value(temp_dict, key_list, value):
            """Set value

            Parameters
            ----------
            temp_dict: dict
                Target dictionary
            key_list: list
                keyword list
            value: any
                value of the item
            """
            if type(key_list) is list:
                for i in range(len(key_list) - 1):
                    k = key_list[i]
                    temp_dict[k] = temp_dict.get(k, {})
                    temp_dict = temp_dict[k]
            temp_dict[key_list[-1]] = value

        filename = filedialog.asksaveasfilename(
            defaultextension=".yaml", filetypes=[("Yaml file", "*.yml *.yaml")]
        )
        if not filename:
            return
        # warning_info
        warning_info = {}
        config_dict = {}
        for tab_index in self.view.microscope_window.tabs():
            microscope_name = self.view.microscope_window.tab(tab_index, "text")
            microscope_tab = self.view.microscope_window.nametowidget(tab_index)
            microscope_dict = {}
            config_dict[microscope_name] = microscope_dict
            for hardware_tab_index in microscope_tab.tabs():
                hardware_name = microscope_tab.tab(hardware_tab_index, "text")
                hardware_tab = microscope_tab.nametowidget(hardware_tab_index)
                hardware_dict = {}
                microscope_dict[
                    hardwares_config_name_dict.get(hardware_name, hardware_name)
                ] = hardware_dict
                for variable_list in hardware_tab.variables_list:
                    if variable_list is None:
                        continue
                    variables, value_dict, ref, format = variable_list
                    if format is None:
                        format = ""
                    temp_dict = hardware_dict
                    if ref is not None:
                        if format.startswith("list"):
                            hardware_dict[ref] = hardware_dict.get(ref, [])
                            temp_dict = {}
                            hardware_dict[ref].append(temp_dict)
                        elif format.startswith("item"):
                            format_list = format.split(";")
                            ref_list = ref.split(";")
                            for i, format in enumerate(format_list):
                                ref = ref_list[i]
                                hardware_dict[ref] = hardware_dict.get(ref, {})
                                temp_dict = hardware_dict[ref]
                                k_idx = format[
                                    format.index("(") + 1 : format.index(",")
                                ].strip()
                                v_idx = format[
                                    format.index(",") + 1 : format.index(")")
                                ].strip()
                                k = variables[k_idx].get()
                                if k.strip() == "":
                                    warning_info[hardware_name] = True
                                    print(
                                        f"Notice: {hardware_name} has an empty value "
                                        f"{ref}! Please double check if it's okay!"
                                    )

                                if k_idx in value_dict:
                                    k = value_dict[k_idx][v]  # noqa
                                v = variables[v_idx].get()
                                if v_idx in value_dict:
                                    v = value_dict[v_idx][v]
                                temp_dict[k] = v
                            continue
                        else:
                            temp_dict = {}
                            hardware_dict[ref] = hardware_dict.get("ref", temp_dict)
                    for k, var in variables.items():
                        try:
                            if k in value_dict:
                                v = value_dict[k][var.get()]
                                # add more information such as file name of a device type
                                if type(v) is tuple:
                                    v = v[0]
                            else:
                                v = var.get()
                        except tk._tkinter.TclError:
                            v = ""
                            print(
                                f"Notice: {hardware_name} has an empty value {k}! "
                                f"Please double check!"
                            )
                            warning_info[hardware_name] = True
                        set_value(temp_dict, k.split("/"), v)

        self.write_to_yaml(config_dict, filename)
        # display warning
        if warning_info:
            messagebox.showwarning(
                title="Configuration",
                message=f"There are empty value(s) with "
                f"{', '.join(warning_info.keys())}"
                f". Please double check!",
            )

    def write_to_yaml(self, config: dict, filename: str) -> None:
        """write yaml file

        Parameters
        ----------
        config: dict
            configuration dictionary
        filename: str
            yaml file name
        """

        def write_func(prefix, config_dict, f):
            for k in config_dict:
                if type(config_dict[k]) == dict:
                    f.write(f"{prefix}{k}:\n")
                    write_func(prefix + " " * 2, config_dict[k], f)
                elif type(config_dict[k]) == list:
                    list_prefix = " "
                    if k != "None":
                        f.write(f"{prefix}{k}:\n")
                        list_prefix = " " * 2
                    for list_item in config_dict[k]:
                        f.write(f"{prefix}{list_prefix}-\n")
                        write_func(prefix + list_prefix * 2, list_item, f)
                elif k != "":
                    f.write(f"{prefix}{k}: {config_dict[k]}\n")

        with open(filename, "w") as f:
            f.write("microscopes:\n")
            write_func("  ", config, f)

    def create_config_window(self, id: int) -> None:
        """Creates the configuration window tabs.

        Parameters
        ----------
        id : int
            The id of the microscope
        """

        tab_name = "Microscope-" + str(id)
        microscope_tab = MicroscopeTab(
            self.view.microscope_window,
            root=self.view.root,
        )
        self.view.microscope_window.tab_list.append(tab_name)
        for hardware_type, widgets in hardwares_dict.items():
            if not widgets:
                continue
            if type(widgets) == dict:
                microscope_tab.create_hardware_tab(hardware_type, widgets)
            else:
                microscope_tab.create_hardware_tab(
                    hardware_type,
                    hardware_widgets=widgets[1],
                    widgets=widgets[2],
                    top_widgets=widgets[0],
                )

        # Adding tabs to self notebook
        self.view.microscope_window.add(
            microscope_tab,
            text=tab_name,
            sticky=tk.NSEW,
        )

    def load_configuration(self) -> None:
        """Load configuration"""

        def get_widget_value(name, value_dict) -> Optional[str]:
            """Get the value from a dict

            Parameters
            ----------
            name: str
                key name
            value_dict: dict
                value dictionary

            Returns
            -------
            value : Optional[str]

                - The value of the key if it exists
                - None if the key does not exist
            """
            value = value_dict
            for key in name.split("/"):
                if key.strip() == "":
                    return value
                value = value.get(key, None)
                if value is None:
                    return None
            return value

        def get_widgets_value(widgets, value_dict):
            """Get all key-value from value_dict, keys are from widgets"""
            temp = {}
            for key in widgets:
                if key == "frame_config":
                    continue
                if widgets[key][1] in ["Button", "Label"]:
                    continue
                value = get_widget_value(key, value_dict)
                # widgets[key][3] is the value mapping dict
                if widgets[key][1] != "Spinbox" and widgets[key][3]:
                    # if the value is not valid, return the last valid value
                    if type(widgets[key][3]) == list:
                        reverse_value_dict = dict(
                            map(lambda v: (v, v), widgets[key][3])
                        )
                    else:
                        reverse_value_dict = dict(
                            map(lambda v: (v[1], v[0]), widgets[key][3].items())
                        )
                    temp[key] = reverse_value_dict.get(
                        value, list(reverse_value_dict.values())[-1]
                    )
                else:
                    temp[key] = value
            return temp

        def build_widgets_value(widgets, value_dict):
            """According to value_dict build values for widgets"""
            if widgets is None or value_dict is None:
                return [None]
            result = []
            ref = ""
            format = ""
            if "frame_config" in widgets:
                ref = widgets["frame_config"].get("ref", "")
                format = widgets["frame_config"].get("format", "")
            if format.startswith("list"):
                if ref != "" and ref.lower() != "none":
                    value_dict = get_widget_value(ref, value_dict)
                if type(value_dict) is not list:
                    return [None]
                for i in range(len(value_dict)):
                    result.append(get_widgets_value(widgets, value_dict[i]))
            elif format.startswith("item"):
                format_list = format.split(";")
                ref_list = ref.split(";")
                for i, format_item in enumerate(format_list):
                    k_idx = format_item[
                        format_item.index("(") + 1 : format_item.index(",")
                    ].strip()
                    v_idx = format_item[
                        format_item.index(",") + 1 : format_item.index(")")
                    ].strip()
                    temp_widget_values = get_widget_value(ref_list[i], value_dict)
                    for j, k in enumerate(temp_widget_values.keys()):
                        if len(result) < j + 1:
                            result.append({k_idx: k, v_idx: temp_widget_values[k]})
                        else:
                            result[j][k_idx] = k
                            result[j][v_idx] = temp_widget_values[k]
            else:
                if ref != "" and ref.lower() != "none":
                    value_dict = get_widget_value(ref, value_dict)
                result.append(get_widgets_value(widgets, value_dict))

            return result

        # ask file name
        file_name = filedialog.askopenfilename(
            defaultextension=".yml", filetypes=[("Yaml file", "*.yml *.yaml")]
        )
        if not file_name:
            return

        # read configuration.yaml
        config_dict = load_yaml_file(file_name)
        if config_dict is None or "microscopes" not in config_dict:
            messagebox.showerror(
                title="Configuration",
                message="It's not a valid configuration.yaml file!",
            )
            return

        self.delete_microscopes()

        for i, microscope_name in enumerate(config_dict["microscopes"].keys()):
            microscope_tab = MicroscopeTab(
                self.view.microscope_window,
                root=self.view.root,
            )
            self.view.microscope_window.add(
                microscope_tab,
                text=microscope_name,
                sticky=tk.NSEW,
            )
            self.view.microscope_window.tab_list.append(microscope_name)

            for hardware_type, widgets in hardwares_dict.items():
                hardware_ref_name = hardwares_config_name_dict[hardware_type]
                # build dictionary values for widgets
                if type(widgets) == dict:
                    try:
                        widgets_value = build_widgets_value(
                            widgets,
                            config_dict["microscopes"][microscope_name][
                                hardware_ref_name
                            ],
                        )
                    except Exception:
                        widgets_value = [None]
                    microscope_tab.create_hardware_tab(
                        hardware_type, widgets, hardware_widgets_value=widgets_value
                    )
                else:
                    try:
                        widgets_value = [
                            build_widgets_value(
                                widgets[1],
                                config_dict["microscopes"][microscope_name][
                                    hardware_ref_name
                                ],
                            ),
                            build_widgets_value(
                                widgets[2],
                                config_dict["microscopes"][microscope_name][
                                    hardware_ref_name
                                ],
                            ),
                        ]
                    except Exception:
                        widgets_value = [[None], [None]]
                    microscope_tab.create_hardware_tab(
                        hardware_type,
                        hardware_widgets=widgets[1],
                        widgets=widgets[2],
                        top_widgets=widgets[0],
                        hardware_widgets_value=widgets_value[0],
                        constants_widgets_value=widgets_value[1],
                    )

    def device_selected(self, event):
        """Handle the event when a device is selected from the dropdown."""
        # # Get the selected device name
        # selected_device_name = self.view.microscope_frame.get()
        # # Find the key in the dictionary that corresponds to the selected value
        # selected_key = next(
        #     key for key, value in device_types.items()
        #     if value == selected_device_name)
        # print(f"Selected Device Key: {selected_key}")
        # print(f"Selected Device Name: {selected_device_name}")
        pass



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard library imports
from pathlib import Path
import os
import tkinter as tk
from tkinter import messagebox

# Third-party imports

# Local application imports
from navigate.config.config import get_navigate_path
from navigate.view.custom_widgets.popup import PopUp
from navigate.tools.file_functions import load_yaml_file, save_yaml_file
from navigate.tools.common_functions import combine_funcs
from navigate.tools.decorators import AcquisitionMode
from navigate.controller.sub_controllers.gui import GUIController
from navigate.view.popups.plugins_popup import PluginsPopup
from navigate.plugins.plugin_manager import PluginFileManager, PluginPackageManager


class PluginsController:
    """Plugins manager in the controller side"""

    def __init__(self, view, parent_controller):
        """Initialize plugins manager class.

        Parameters
        ----------
        view: object
            tkinter frame object.
        parent_controller: object
            navigate controller.
        """
        #: object: tkinter frame object.
        self.view = view

        #: object: navigate controller.
        self.parent_controller = parent_controller

        #: dict: installed plugins with GUI
        self.plugins_dict = {}

    def populate_experiment_setting(self):
        """populate experiment values to plugin GUI"""
        for plugin_name in self.plugins_dict:
            try:
                self.plugins_dict[plugin_name].populate_experiment_setting()
            except Exception:
                pass

    def load_plugins(self):
        """Load plugins"""
        # load through files
        plugins_path = os.path.join(
            Path(__file__).resolve().parent.parent.parent, "plugins"
        )
        plugins_config_path = os.path.join(
            get_navigate_path(), "config", "plugins_config.yml"
        )
        plugin_file_manager = PluginFileManager(plugins_path, plugins_config_path)
        self.load_plugins_through_manager(plugin_file_manager)
        self.load_plugins_through_manager(PluginPackageManager)

    def load_plugins_through_manager(self, plugin_manager):
        """Load plugins through plugin manager

        Parameters
        ----------
        plugin_manager : object
            PluginManager object
        """
        plugins = plugin_manager.get_plugins()

        for plugin_name, plugin_ref in plugins.items():

            # read "plugin_config.yml"
            plugin_config = plugin_manager.load_config(plugin_ref)
            if plugin_config is None:
                continue
            plugin_display_name = plugin_config.get("name", plugin_name)

            plugin_frame = plugin_manager.load_view(plugin_ref, plugin_display_name)
            plugin_controller = plugin_manager.load_controller(
                plugin_ref, plugin_display_name
            )

            if plugin_frame and plugin_controller:
                if plugin_config["view"] == "Popup":
                    # menu
                    self.parent_controller.view.menubar.menu_plugins.add_command(
                        label=plugin_name,
                        command=self.build_popup_window(
                            plugin_name, plugin_frame, plugin_controller
                        ),
                    )
                else:
                    self.build_tab_window(plugin_name, plugin_frame, plugin_controller)
            # feature
            plugin_manager.load_features(plugin_ref)

            # acquisition mode
            acquisition_modes = plugin_config.get("acquisition_modes", [])
            plugin_manager.load_acquisition_modes(
                plugin_ref,
                acquisition_modes,
                self.register_acquisition_mode,
            )

    def build_tab_window(self, plugin_name, frame, controller):
        """Build tab for a plugin

        Parameters
        ----------
        plugin_name : str
            plugin name.
        frame: object
            navigate frame object
        controller: object
            Controller Class
        """
        try:
            plugin_frame = frame(self.view.settings)
            self.view.settings.add(plugin_frame, text=plugin_name, sticky=tk.NSEW)
            plugin_controller = controller(plugin_frame, self.parent_controller)
            self.parent_controller.register_event_listeners(
                getattr(plugin_controller, "custom_events", {})
            )
            controller_name = (
                "__plugin" + "_".join(plugin_name.lower().split()) + "_controller"
            )
            self.plugins_dict[controller_name] = plugin_controller
        except Exception as e:
            messagebox.showwarning(
                title="Warning",
                message=(
                    f"Plugin {plugin_name} has something went wrong."
                    f"Error: {e}"
                    f"Please make sure the plugin works correctly or uninstall it!"
                ),
            )

    def build_popup_window(self, plugin_name, frame, controller):
        """Build popup window for a plugin

        Parameters
        ----------
        plugin_name : str
            plugin name.
        frame: object
            navigate frame object
        controller: object
            navigate controller
        """
        controller_name = (
            "__plugin" + "_".join(plugin_name.lower().split()) + "_controller"
        )

        def func(*args, **kwargs):
            """Function to build popup window for a plugin

            Parameters
            ----------
            args: list
                arguments.
            kwargs: dict
                keyword arguments.
            """
            if controller_name in self.plugins_dict:
                self.plugins_dict[controller_name].popup.deiconify()
                return
            popup = PopUp(self.view, plugin_name, "+320+180", transient=False)
            popup.configure(bg="white")
            content_frame = popup.get_frame()
            plugin_frame = frame(content_frame)
            plugin_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=10, pady=10)

            plugin_controller = controller(plugin_frame, self.parent_controller)
            self.parent_controller.register_event_listeners(
                getattr(plugin_controller, "custom_events", {})
            )

            plugin_controller.popup = popup
            popup.deiconify()
            self.plugins_dict[controller_name] = plugin_controller

            popup.protocol(
                "WM_DELETE_WINDOW",
                combine_funcs(
                    popup.dismiss, lambda: self.plugins_dict.pop(controller_name)
                ),
            )

        def func_with_wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except Exception:
                messagebox.showwarning(
                    title="Warning",
                    message=(
                        f"Plugin {plugin_name} has something went wrong."
                        "Please make sure the plugin works correctly or uninstall it "
                        "from navigate!"
                    ),
                )

        return func_with_wrapper

    def register_acquisition_mode(self, acquisition_mode_name, module):
        """Register acquisition mode

        Parameters
        ----------
        acquisition_mode_name : str
            The name of an acquisition mode
        module : module
            acquisition mode module
        """
        if not module:
            return
        acquisition_mode = [
            m for m in dir(module) if isinstance(getattr(module, m), AcquisitionMode)
        ]
        if acquisition_mode:
            self.parent_controller.add_acquisition_mode(
                acquisition_mode_name,
                getattr(module, acquisition_mode[0]),
            )


class UninstallPluginController(GUIController):
    """Uninstall plugin controller"""

    def __init__(self, view, parent_controller):
        """Initialize uninstall plugin controller class.

        Parameters
        ----------
        view: object
            tkinter frame object.
        parent_controller: object
            navigate controller.
        """
        super().__init__(view, parent_controller)

        #: str: plugin config path.
        self.plugin_config_path = os.path.join(get_navigate_path(), "config")

        #: PluginsPopup: popup window object.
        self.popup = PluginsPopup(view)
        self.popup.uninstall_btn.config(command=self.uninstall_plugins)
        self.popup.popup.protocol("WM_DELETE_WINDOW", self.exit_func)
        self.refresh_plugins()

    def showup(self):
        """Show up the popup window"""
        self.refresh_plugins()
        self.popup.popup.showup()

    def exit_func(self):
        """Exit"""
        self.popup.popup.dismiss()
        delattr(self.parent_controller, "uninstall_plugin_controller")

    def refresh_plugins(self):
        """Show installed plugins"""
        self.plugin_config = load_yaml_file(
            os.path.join(self.plugin_config_path, "plugins_config.yml")
        )

        if self.plugin_config is None:
            self.plugin_config = {}

        self.popup.build_widgets(self.plugin_config)

    def uninstall_plugins(self, *args):
        """Uninstall plugins.

        Parameters
        ----------
        args: list
            arguments.
        """
        feature_lists_path = get_navigate_path() + "/feature_lists"
        features = os.listdir(feature_lists_path)
        feature_config = {}
        for feature_name in features:
            if feature_name.endswith(".yml") and feature_name != "__sequence.yml":
                feature_content = load_yaml_file(
                    os.path.join(feature_lists_path, feature_name)
                )
                if feature_content and feature_content.get("filename", None):
                    feature_config[feature_name] = feature_content["filename"]
        flag = False
        for var in self.popup.variables:
            if var.get():
                # remove the feature list if a deleted plugin has any.
                for feature_name in feature_config:
                    if feature_config[feature_name].startswith(
                        self.plugin_config[var.get()]
                    ):
                        os.remove(f"{feature_lists_path}/{feature_name}")
                self.plugin_config.pop(var.get())
                flag = True
        if flag:
            save_yaml_file(
                self.plugin_config_path, self.plugin_config, "plugins_config.yml"
            )
            messagebox.showwarning(
                title="Navigate",
                message="Plugins are uninstalled! Please restart Navigate!",
            )
            self.showup()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.
from typing import Optional, Dict, Any

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports

# Third-party imports
import zarr
import numpy.typing as npt
import zarr.storage

# Local application imports
from .pyramidal_data_source import PyramidalDataSource
from ..metadata_sources.zarr_metadata import OMEZarrMetadata

GROUP_PREFIX = "p"


class OMEZarrDataSource(PyramidalDataSource):
    """OME-Zarr data source.

    This class implements an OME-Zarr image data source using the Zarr v2 format.
    """

    def __init__(self, file_name: str = None, mode: str = "w") -> None:
        """Initialize the OME-Zarr data source.

        Parameters
        ----------
        file_name : str
            Name of the file to open.
        mode : str
            Mode to open the file in.
        """

        #: OMEZarrMetadata: Metadata object for the OME-Zarr data source.
        self.metadata = OMEZarrMetadata()
        self.__store = None
        self._current_position = -1

        super().__init__(file_name=file_name, mode=mode)

    def get_slice(self, x, y, c, z=0, t=0, p=0, subdiv=0) -> npt.ArrayLike:
        """Get a 3D slice of the dataset for a single c, t, p, subdiv.

        Parameters
        ----------
        x : int or slice
            x indices to grab
        y : int or slice
            y indices to grab
        c : int
            Single channel
        z : int or slice
            z indices to grab
        t : int
            Single timepoint
        p : int
            Single position
        subdiv : int
            Subdivision of the dataset to index along

        Returns
        -------
        npt.ArrayLike
            3D (z, y, x) slice of data set
        """
        dataset_name = f"{GROUP_PREFIX}{p}_{subdiv}"
        return self.image[dataset_name][t, c, z, y, x]

    def setup(self):
        """Set up the Zarr writer."""
        # Use FSStore as a universal backend
        self.__store = zarr.storage.FSStore(self.file_name, mode=self.mode)

        #: zarr.group: Zarr group object for the image data source.
        self.image = zarr.group(store=self.__store, overwrite=True)
        self._current_position = -1

    def new_position(self, pos, view):
        """Create new arrays on the fly for each position in self.positions.

        Parameters
        ----------
        pos : int
            Position index
        view : str
            View name
        """
        name = f"{GROUP_PREFIX}{pos}"
        paths = []
        # Create the subdivisions...
        for si, zyx_shape in enumerate(self.shapes):
            shape = tuple([self.shape_t, self.shape_c] + list(zyx_shape))
            setup = f"{name}_{si}"
            arr = self.image.create(
                name=setup,
                shape=shape,
                chunks=(1,) * len(shape[:-2]) + shape[-2:],
                dtype=self.dtype,
            )
            # xarray multidim
            paths.append(arr.path)
            arr.attrs["_ARRAY_DIMENSIONS"] = shape

        # Append setup to multiscales
        scales = self.image.attrs.get("multiscales", [])
        scales.append(
            self.metadata.multiscales_dict(name, paths, self.resolutions, view)
        )
        self.image.attrs["multiscales"] = scales

    def write(self, data: npt.ArrayLike, **kw) -> None:
        """Writes 2D image to the data source.

        Parameters
        ----------
        data : npt.ArrayLike
            2D image data
        kw : dict
            Keyword arguments
        """
        #: str: Mode of the data source.
        self.mode = "w"

        if self.__store is None:
            self.setup()

        c, z, t, p = self._cztp_indices(
            self._current_frame, self.metadata.per_stack
        )  # find current channel

        if self._current_position != p:
            self._current_position = p
            if len(kw) > 0:
                self.new_position(p, kw)
            else:
                self.new_position(p)

        # TODO: Make sure this also functions.
        for ri, res in enumerate(self.resolutions):
            dx, dy, dz = res
            dataset_name = f"{GROUP_PREFIX}{p}_{ri}"
            zs = min(z // dz, self.shapes[ri, 0] - 1)
            self.image[dataset_name][t, c, zs, ...] = data[::dy, ::dx].astype(
                self.dtype
            )

        self._current_frame += 1

    def read(self) -> None:
        """Reads data from the image file."""
        self.mode = "r"
        self.__store = zarr.storage.FSStore(self.file_name, mode=self.mode)
        self.image = zarr.group(store=self.__store)
        # TODO: parse the image metadata
        self.get_shape_from_metadata()

    def close(self) -> None:
        """Close the image file."""
        if self._closed:
            if self.__store is not None:
                self.__store = None
            return
        self._check_shape(self._current_frame - 1, self.metadata.per_stack)
        self.__store.close()
        self._closed = True
        self.__store = None
